"""
This module contains the Slisemap class.
"""

from copy import copy
from os import PathLike
from typing import Any, BinaryIO, Callable, Dict, List, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.patches import Patch

from slisemap.escape import escape_neighbourhood
from slisemap.local_models import linear_regression, linear_regression_loss
from slisemap.loss import make_loss, make_marginal_loss, softmax_kernel
from slisemap.utils import (
    LBFGS,
    CheckConvergence,
    PCA_rotation,
    _assert,
    _tonp,
    _warn,
    global_model,
    varimax,
)


class Slisemap:
    """Slisemap: Combine local explanations with dimensionality reduction.

    Args:
        X (Union[np.ndarray, torch.Tensor]): Data matrix. Note that the data is assumed to be normalised.
        y (Union[np.ndarray, torch.Tensor]): Target vector or matrix.
        radius (float, optional): The radius of the embedding Z. Defaults to 3.5.
        d (int, optional): The number of embedding dimensions. Defaults to 2.
        lasso (Optional[float], optional): Lasso regularisation coefficient. Defaults to 0.0.
        ridge (Optional[float], optional): Ridge regularisation coefficient. Defaults to 0.0.
        z_norm (float, optional): Z normalisation regularisation coefficient. Defaults to 0.01.
        intercept (bool, optional): Should an intercept term be added to self.X. Defaults to True.
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor], optional): Local model prediction function. Defaults to linear_regression.
        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor], optional): Local model loss function. Defaults to linear_regression_loss.
        coefficients (Union[None, int, Callable[[torch.Tensor, torch.Tensor], int]], optional): The number of local model coefficients or a function: `(X,Y)->coefficients`. Defaults to self.X.shape[1] * self.Y.shape[1].
        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor], optional): Distance function. Defaults to torch.cdist (Euclidean distance).
        kernel (Callable[[torch.Tensor], torch.Tensor], optional): Kernel function. Defaults to softmax_kernel.
        B0 (Union[None, np.ndarray, torch.Tensor], optional): Initial value for B (random if None). Defaults to None.
        Z0 (Union[None, np.ndarray, torch.Tensor], optional): Initial value for Z (PCA if None). Defaults to None.
        dtype (torch.dtype, optional): Floating type. Defaults to torch.float32.
        jit (bool, optional): Just-In-Time compile the loss function for increased performance (see `torch.jit.trace` for caveats). Defaults to True.
        random_state (Optional[int], optional): Set an explicit seed for the random number generator (i.e. `torch.manual_seed`). Defaults to None.
        cuda (Optional[bool], optional): Use cuda if available (defaults to true if the data is large enough). Defaults to None.
    """

    # Make Python faster and safer by not creating a Slisemap.__dict__
    __slots__ = (
        "_X",
        "_Y",
        "_Z",
        "_Z0",
        "_B",
        "_B0",
        "_radius",
        "_lasso",
        "_ridge",
        "_z_norm",
        "_intercept",
        "_local_model",
        "_local_loss",
        "_loss",
        "_distance",
        "_kernel",
        "_jit",
        "_random_state",
        "_rs0",
    )

    def __init__(
        self,
        X: Union[np.ndarray, torch.Tensor],
        y: Union[np.ndarray, torch.Tensor],
        radius: float = 3.5,
        d: int = 2,
        lasso: Optional[float] = None,
        ridge: Optional[float] = None,
        z_norm: float = 0.01,
        intercept: bool = True,
        local_model: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = linear_regression,
        local_loss: Callable[
            [torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor
        ] = linear_regression_loss,
        coefficients: Union[
            None, int, Callable[[torch.Tensor, torch.Tensor], int]
        ] = None,
        distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = torch.cdist,
        kernel: Callable[[torch.Tensor], torch.Tensor] = softmax_kernel,
        B0: Union[None, np.ndarray, torch.Tensor] = None,
        Z0: Union[None, np.ndarray, torch.Tensor] = None,
        dtype: torch.dtype = torch.float32,
        jit: bool = True,
        random_state: Optional[int] = None,
        cuda: Optional[bool] = None,
    ):
        if lasso is None and ridge is None:
            _warn(
                "Consider using regularisation (lasso/l1 and ridge/l2 regularisation is built-in, via the parameters `lasso` and `ridge`). "
                + "Regularisation is important for handling small neighbourhoods, and also makes the local models more local. "
                + "Set `lasso=0` to disable this warning (if no regularisation is really desired).",
                Slisemap,
            )
        self.lasso = 0.0 if lasso is None else lasso
        self.ridge = 0.0 if ridge is None else ridge
        self.kernel = kernel
        self.distance = distance
        self.local_model = local_model
        self.local_loss = local_loss
        self.z_norm = z_norm
        self.radius = radius
        self._intercept = intercept
        self._jit = jit
        self._rs0 = random_state

        if cuda is None:
            scale = (
                X.shape[0] ** 2 * X.shape[1] * (1 if len(y.shape) < 2 else y.shape[1])
            )
            cuda = scale > 1_000_000 and torch.cuda.is_available()
        tensorargs = {
            "device": torch.device("cuda" if cuda else "cpu"),
            "dtype": dtype,
        }

        self._X = torch.as_tensor(X, **tensorargs)
        if intercept:
            self._X = torch.cat((self._X, torch.ones_like(self._X[:, :1])), 1)
        n, m = self._X.shape

        self.random_state = random_state

        self._Y = torch.as_tensor(y, **tensorargs)
        _assert(
            self._Y.shape[0] == n,
            f"The length of y must match X: {self._Y.shape[0]} != {n}",
        )
        if len(self._Y.shape) == 1:
            self._Y = self._Y[:, None]

        if Z0 is None:
            self._Z0 = self._X @ PCA_rotation(self._X, min(m, d))
            if d > m:
                _warn(
                    "The number of embedding dimensions is larger than the number of data dimensions",
                    Slisemap,
                )
                Z0fill = torch.zeros(size=[n, d - m], **tensorargs)
                self._Z0 = torch.cat((self._Z0, Z0fill), 1)
        else:
            self._Z0 = torch.as_tensor(Z0, **tensorargs)
            _assert(
                self._Z0.shape == (n, d),
                f"Z0 has the wrong shape: {self._Z0.shape} != ({n}, {d})",
            )
        if radius > 0:
            norm = 1 / (torch.sqrt(torch.sum(self._Z0**2) / self._Z0.shape[0]) + 1e-8)
            self._Z0 = self._Z0 * norm
        self._Z = self._Z0.detach().clone()

        if callable(coefficients):
            coefficients = coefficients(self._X, self._Y)
        if B0 is None:
            if coefficients is None:
                coefficients = m * self.p
            B0 = global_model(
                X=self._X,
                Y=self._Y,
                local_model=self.local_model,
                local_loss=self.local_loss,
                coefficients=coefficients,
                lasso=self.lasso,
                ridge=self.ridge,
            )
            if torch.all(torch.isfinite(B0)):
                self._B0 = B0.expand((n, coefficients))
            else:
                _warn(
                    "Optimising a global model as initialisation resulted in non-finite values. Consider using stronger regularisation (increase `lasso` or `ridge`).",
                    Slisemap,
                )
                self._B0 = torch.zeros((n, coefficients), **tensorargs)
        else:
            self._B0 = torch.as_tensor(B0, **tensorargs)
            if coefficients is None:
                _assert(len(B0.shape) > 1, "B0 must have more than one dimension")
                coefficients = B0.shape[1]
            _assert(
                self._B0.shape == (n, coefficients),
                f"B0 has the wrong shape: {self._B0.shape} != ({n}, {coefficients})",
            )
        self._B = self._B0.detach().clone()

    def restore(self):
        """Reset B and Z (and random_state) to their initial values B0 and Z0 (and _rs0)."""
        self._Z = self._Z0.clone().detach()
        self._B = self._B0.clone().detach()
        self.random_state = self._rs0

    @property
    def n(self) -> int:
        """The number of data items."""
        return self._X.shape[0]

    @property
    def m(self) -> int:
        """The number of variables (including potential intercept)."""
        return self._X.shape[1]

    @property
    def p(self) -> int:
        """The number of target variables (i.e. the number of classes)."""
        return self._Y.shape[-1]

    @property
    def d(self) -> int:
        """The number of embedding dimensions."""
        return self._Z.shape[1]

    @d.setter
    def d(self, value: int):
        _assert(value > 0, "The number of embedding dimensions must be positive")
        if self.d != value:
            self._loss = None  # invalidate cached loss function
            self._Z = self._Z.detach()
            if self.d > value:
                self._Z = self._Z @ PCA_rotation(self._Z, value)
            else:
                zn = [self._Z, torch.zeros((self.n, value - self.d), **self.tensorargs)]
                self._Z = torch.concat(zn, 1)

    @property
    def coefficients(self) -> int:
        """The number of local model coefficients."""
        return self._B.shape[1]

    @property
    def intercept(self) -> bool:
        """Is an intercept column added to the data?"""
        return self._intercept

    @property
    def radius(self) -> float:
        """The radius of the embedding."""
        return self._radius

    @radius.setter
    def radius(self, value: float):
        _assert(value >= 0, "radius must not be negative")
        self._radius = value
        self._loss = None  # invalidate cached loss function

    @property
    def lasso(self) -> float:
        """Lasso regularisation strength."""
        return self._lasso

    @lasso.setter
    def lasso(self, value: float):
        _assert(value >= 0, "lasso must not be negative")
        self._lasso = value
        self._loss = None  # invalidate cached loss function

    @property
    def ridge(self) -> float:
        """Ridge regularisation strength."""
        return self._ridge

    @ridge.setter
    def ridge(self, value: float):
        _assert(value >= 0, "ridge must not be negative")
        self._ridge = value
        self._loss = None  # invalidate cached loss function

    @property
    def z_norm(self) -> float:
        """Z normalisation regularisation strength."""
        return self._z_norm

    @z_norm.setter
    def z_norm(self, value: float):
        _assert(value >= 0, "z_norm must not be negative")
        self._z_norm = value
        self._loss = None  # invalidate cached loss function

    @property
    def local_model(self) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Local model prediction function. Takes in X[n, m] and B[n, coefficients], and returns Ytilde[n, n, p]."""
        return self._local_model

    @local_model.setter
    def local_model(self, value: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]):
        _assert(callable(value), "local_model must be callable")
        self._local_model = value
        self._loss = None  # invalidate cached loss function

    @property
    def local_loss(
        self,
    ) -> Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]:
        """Local model loss function. Takes in Ytilde[n, n, p], Y[n, p], and B[n, coefficients], and returns L[n, n]"""
        return self._local_loss

    @local_loss.setter
    def local_loss(
        self, value: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]
    ):
        _assert(callable(value), "local_loss must be callable")
        self._local_loss = value
        self._loss = None  # invalidate cached loss function

    @property
    def distance(self) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
        """Distance function. Takes in Z[n1, d] and Z[n2, d], and returns D[n1, n2]"""
        return self._distance

    @distance.setter
    def distance(self, value: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]):
        _assert(callable(value), "distance must be callable")
        self._distance = value
        self._loss = None  # invalidate cached loss function

    @property
    def kernel(self) -> Callable[[torch.Tensor], torch.Tensor]:
        """Kernel function. Takes in D[n, n] and returns W[n, n]"""
        return self._kernel

    @kernel.setter
    def kernel(self, value: Callable[[torch.Tensor], torch.Tensor]):
        _assert(callable(value), "kernel must be callable")
        self._kernel = value
        self._loss = None  # invalidate cached loss function

    @property
    def jit(self) -> bool:
        """Just-In-Time compile the loss function."""
        return self._jit

    @jit.setter
    def jit(self, value: bool):
        self._jit = value
        self._loss = None  # invalidate cached loss function

    def _set_random_state(self, value: Optional[int]):
        """Set the seed for the random number generator specific for this object (None reverts to the global torch PRNG)."""
        if value is None:
            self._random_state = None
        else:
            if self.X.device.type == "cpu":
                self._random_state = torch.random.manual_seed(value)
            elif self.X.device.type == "cuda":
                gen = torch.cuda.default_generators[self.X.device.index]
                self._random_state = gen.manual_seed(value)
            else:
                _warn(
                    Slisemap._set_random_state,
                    "Unknown device, setting the global seed insted",
                )
                torch.random.manual_seed(value)
                self._random_state = None

    random_state = property(fset=_set_random_state)
    del _set_random_state

    @property
    def X(self) -> torch.Tensor:
        """Data matrix as a `torch.Tensor` (use `Slisemap.get_X()` for options)."""
        return self._X

    @property
    def Y(self) -> torch.Tensor:
        """Target matrix as a `torch.Tensor` (use `Slisemap.get_Y()` for options)."""
        return self._Y

    @property
    def Z(self) -> torch.Tensor:
        """Normalised embedding matrix as a `torch.Tensor` (use `Slisemap.get_Z()` for options)."""
        return self._Z

    @property
    def B(self) -> torch.Tensor:
        """Coefficient matrix for the local models as a `torch.Tensor` (use `Slisemap.get_B()` for options)."""
        return self._B

    @property
    def tensorargs(self) -> Dict[str, Any]:
        """When creating a new `torch.Tensor` add these keyword arguments to match the `dtype` and `device` of this Slisemap object."""
        return dict(device=self._X.device, dtype=self._X.dtype)

    def cuda(self, **kwargs):
        """Move the tensors to CUDA memory (and run the calculations there).
        Note that this resets the random state.

        Args:
            **kwargs: Optional arguments to `torch.Tensor.cuda`
        """
        X = self._X.cuda(**kwargs)
        self._X = X
        self._Y = self._Y.cuda(**kwargs)
        self._Z = self._Z.detach().cuda(**kwargs)
        self._B = self._B.detach().cuda(**kwargs)
        self.random_state = self._rs0
        self._loss = None  # invalidate cached loss function

    def cpu(self, **kwargs):
        """Move the tensors to CPU memory (and run the calculations there).
        Note that this resets the random state.

        Args:
            **kwargs: Optional arguments to `torch.Tensor.cpu`
        """
        X = self._X.cpu(**kwargs)
        self._X = X
        self._Y = self._Y.cpu(**kwargs)
        self._Z = self._Z.detach().cpu(**kwargs)
        self._B = self._B.detach().cpu(**kwargs)
        self.random_state = self._rs0
        self._loss = None  # invalidate cached loss function

    def get_loss_fn(
        self, individual: bool = False
    ) -> Callable[
        [torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor
    ]:
        """Returns the slisemap loss function.
        This function JITs and caches the loss function for efficiency.

        Args:
            individual (bool, optional): Make a loss function for individual losses. Defaults to False.

        Returns:
            Callable[[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]: Loss function: (X, Y, B, Z) -> loss.
        """
        if individual:
            return make_loss(
                local_model=self.local_model,
                local_loss=self.local_loss,
                distance=self.distance,
                kernel=self.kernel,
                radius=self.radius,
                lasso=self.lasso,
                ridge=self.ridge,
                z_norm=0.0,
                individual=True,
            )
        if self._loss is None:
            # Caching the loss function
            self._loss = make_loss(
                local_model=self.local_model,
                local_loss=self.local_loss,
                distance=self.distance,
                kernel=self.kernel,
                radius=self.radius,
                lasso=self.lasso,
                ridge=self.ridge,
                z_norm=self.z_norm,
            )
            # JITting the loss function improves the performance
            if self._jit:
                self._loss = torch.jit.trace(
                    self._loss, (self._X[:1], self._Y[:1], self._B[:1], self._Z[:1])
                )
        return self._loss

    def _as_new_X(
        self, X: Union[None, np.ndarray, torch.Tensor] = None
    ) -> torch.Tensor:
        if X is None:
            return self._X
        tensorargs = self.tensorargs
        X = torch.atleast_2d(torch.as_tensor(X, **tensorargs))
        if self._intercept:
            X = torch.cat([X, torch.ones((X.shape[0], 1), **tensorargs)], 1)
        _assert(
            X.shape[1] == self.m,
            f"X has the wrong shape {X.shape[1]} != {self.m}",
        )
        return X

    def _as_new_Y(
        self, Y: Union[None, float, np.ndarray, torch.Tensor] = None, n: int = -1
    ) -> torch.Tensor:
        if Y is None:
            return self._Y
        Y = torch.as_tensor(Y, **self.tensorargs)
        if len(Y.shape) < 2:
            Y = torch.reshape(Y, (n, self.p))
        n = Y.shape[0]
        _assert(
            Y.shape == (n, self.p),
            f"Y has the wrong shape {Y.shape} != {(n, self.p)}",
        )
        return Y

    def get_Z(
        self, scale: bool = True, rotate: bool = False, numpy: bool = True
    ) -> Union[np.ndarray, torch.Tensor]:
        """Get the Z matrix

        Args:
            scaled (bool, optional): Scale the returned Z to match self.radius. Defaults to True.
            rotate (bool, optional): Rotate the returned Z so that the first dimension is the major axis. Defaults to False.
            numpy (bool, optional): Return the matrix as a numpy (True) or pytorch (False) matrix. Defaults to True.

        Returns:
            Union[np.ndarray, torch.Tensor]: The Z matrix
        """
        if scale:
            if self.radius > 0:
                Zss = torch.sum(self._Z**2)
                Z = self._Z * (
                    self.radius / (torch.sqrt(Zss / self._Z.shape[0]) + 1e-8)
                )
            else:
                Z = self._Z - self._Z.mean(dim=0, keepdim=True)
        else:
            Z = self._Z
        if rotate:
            Z = varimax(Z)
        if numpy:
            return Z.cpu().detach().numpy()
        else:
            return Z.detach().clone()

    def get_B(self, numpy: bool = True) -> Union[np.ndarray, torch.Tensor]:
        """Get the B matrix

        Args:
            numpy (bool, optional): Return the matrix as a numpy (True) or pytorch (False) matrix. Defaults to True.

        Returns:
            Union[np.ndarray, torch.Tensor]: The B matrix
        """
        return _tonp(self._B) if numpy else self._B.detach().clone()

    def get_D(self, numpy: bool = True) -> Union[np.ndarray, torch.Tensor]:
        """Get the embedding distance matrix

        Args:
            numpy (bool, optional): Return the matrix as a numpy (True) or pytorch (False) matrix. Defaults to True.

        Returns:
            Union[np.ndarray, torch.Tensor]: The D matrix
        """
        Z = self.get_Z(rotate=False, numpy=False)
        D = self._distance(Z, Z)
        return _tonp(D) if numpy else D.detach()

    def get_L(
        self,
        X: Union[None, np.ndarray, torch.Tensor] = None,
        Y: Union[None, float, np.ndarray, torch.Tensor] = None,
        numpy: bool = True,
    ) -> Union[np.ndarray, torch.Tensor]:
        """Get the loss matrix: [B.shape[0], X.shape[0]].

        Args:
            X (Union[None, np.ndarray, torch.Tensor], optional): Optional replacement for the training X. Defaults to None.
            Y (Union[None, float, np.ndarray, torch.Tensor], optional): Optional replacement for the training Y. Defaults to None.
            numpy (bool, optional): Return the matrix as a numpy (True) or pytorch (False) matrix. Defaults to True.

        Returns:
            Union[np.ndarray, torch.Tensor]: The L matrix
        """
        X = self._as_new_X(X)
        Y = self._as_new_Y(Y, X.shape[0])
        L = self.local_loss(self.local_model(X, self._B), Y, self._B)
        return _tonp(L) if numpy else L.detach()

    def get_W(self, numpy: bool = True) -> Union[np.ndarray, torch.Tensor]:
        """Get the weight matrix

        Args:
            numpy (bool, optional): Return the matrix as a numpy.ndarray instead of a torch.Tensor. Defaults to True.

        Returns:
            Union[np.ndarray, torch.Tensor]: The W matrix
        """
        W = self.kernel(self.get_D(numpy=False))
        return _tonp(W) if numpy else W.detach()

    def get_X(
        self, numpy: bool = True, intercept: bool = True
    ) -> Union[np.ndarray, torch.Tensor]:
        """Get the data matrix

        Args:
            numpy (bool, optional): Return the matrix as a numpy.ndarray instead of a torch.Tensor. Defaults to True.
            intercept (bool, optional): Include the intercept column (if `self.intercept == True`). Defaults to True.

        Returns:
            Union[np.ndarray, torch.Tensor]: The X matrix
        """
        X = self._X if intercept or not self._intercept else self._X[:, :-1]
        return _tonp(X) if numpy else X.detach().clone()

    def get_Y(
        self, numpy: bool = True, ravel: bool = False
    ) -> Union[np.ndarray, torch.Tensor]:
        """Get the target matrix

        Args:
            numpy (bool, optional): Return the matrix as a numpy.ndarray instead of a torch.Tensor. Defaults to True.
            ravel (bool, optional): Remove the second dimension if it is singular (i.e. turn it into a vector). Defaults to False.

        Returns:
            Union[np.ndarray, torch.Tensor]: The Y matrix
        """
        Y = self._Y.ravel() if ravel else self._Y
        return _tonp(Y) if numpy else Y.detach().clone()

    def value(
        self, individual: bool = False, numpy: bool = True
    ) -> Union[float, np.ndarray, torch.Tensor]:
        """Calculate the loss value.

        Args:
            individual (bool, optional): Give loss individual loss values for the data points. Defaults to False.

        Returns:
            Union[float, np.ndarray]: loss value(s).
        """
        loss = self.get_loss_fn(individual)
        loss = loss(X=self._X, Y=self._Y, B=self._B, Z=self._Z)
        if individual:
            return _tonp(loss) if numpy else loss.detach()
        else:
            return loss.cpu().detach().item() if numpy else loss.detach()

    def entropy(
        self, aggregate: bool = True, numpy: bool = True
    ) -> Union[float, np.ndarray, torch.Tensor]:
        """Compute row-wise entropy of the W matrix induced by Z.

        Args:
            aggregate (bool, optional): Aggregate the row-wise entropies into one scalar. Defaults to True.
            numpy (bool, optional): return as a numpy.ndarray or float instead of torch.Tensor. Defaults to True.

        Returns:
            Union[float, np.ndarray, torch.Tensor]: entropy
        """
        W = self.get_W(False)
        entropy = -(W * W.log()).sum(dim=1)
        if aggregate:
            entropy = (entropy.mean().exp() / self.n).detach()
            return entropy.cpu().item() if numpy else entropy
        else:
            return _tonp(entropy) if numpy else entropy.detach()

    def lbfgs(self, max_iter: int = 500, *, only_B: bool = False, **kwargs) -> float:

        """Optimise Slisemap using LBFGS.

        Args:
            max_iter (int, optional): Maximum number of LBFGS iterations. Defaults to 500.
            only_B (bool, optional): Only optimise B. Defaults to False.
            **kwargs: Optional keyword arguments to LBFGS.

        Returns:
            float: The loss value.
        """
        Z = self._Z.detach().clone().requires_grad_(True)
        B = self._B.detach().clone().requires_grad_(True)

        loss_ = self.get_loss_fn()
        loss_fn = lambda: loss_(self._X, self._Y, B, Z)
        LBFGS(loss_fn, [B] if only_B else [Z, B], max_iter=max_iter, **kwargs)

        if np.isnan(loss_fn().cpu().detach().item()):
            _warn(
                "An LBFGS optimisation resulted in `nan` (try strengthening the regularisation or reducing the radius)",
                Slisemap.lbfgs,
            )
            # Some datasets (with logistic local models) are initially numerically unstable.
            # Just running LBFGS for one iteration seems to avoid those issues.
            Z = self._Z.detach().requires_grad_(True)
            B = self._B.detach().requires_grad_(True)
            LBFGS(loss_fn, [B] if only_B else [Z, B], max_iter=1, **kwargs)

        self._Z = Z.detach()
        self._B = B.detach()
        self._normalise()

        return self.value()

    def escape(
        self,
        force_move: bool = True,
        escape_fn: Callable = escape_neighbourhood,
    ):
        """Try to escape a local optimum by moving the items (embedding and local model) to the neighbourhoods best suited for them.
        This is done by finding another item (in the optimal neighbourhood) and copying its values for Z and B.

        Args:
            force_move (bool, optional): Do not allow the items to pair with themselves. Defaults to True.
            escape_fn (Callable, optional): Escape function (escape_neighbourhood/escape_greedy/escape_marginal). Defaults to escape_neighbourhood.
        """
        self._B, self._Z = escape_fn(
            X=self._X,
            Y=self._Y,
            B=self._B,
            Z=self._Z,
            local_model=self.local_model,
            local_loss=self.local_loss,
            distance=self.distance,
            kernel=self.kernel,
            radius=self.radius,
            force_move=force_move,
            jit=self.jit,
        )
        self._normalise()

    def _normalise(self):
        """Normalise Z."""
        if self.radius > 0:
            self._Z *= 1 / (torch.sqrt(torch.sum(self._Z**2) / self.n) + 1e-8)

    def optimise(
        self,
        patience: int = 3,
        max_escapes: int = 100,
        max_iter: int = 500,
        escape_fn: Callable = escape_neighbourhood,
        verbose: bool = False,
        noise: float = 1e-4,
        **kwargs,
    ) -> float:
        """Optimise Slisemap by alternating between Slisemap.lbfgs and Slisemap.escape until convergence.

        Args:
            patience (int, optional): Number of escapes without improvement before stopping. Defaults to 3.
            max_escapes (int, optional): Maximum number of escapes. Defaults to 100.
            max_iter (int, optional): Maximum number of LBFGS iterations per round. Defaults to 500.
            escape_fn (Callable, optional): Escape function (escape_neighbourhood/escape_greedy/escape_marginal). Defaults to escape_neighbourhood.
            verbose (bool, optional): Print status messages. Defaults to False.
            noise (float, optional): Scale of the noise used to avoid disappearing dimensions in the embedding after an escape (gradient based optimisers cannot recover them). Defaults to 1e-4.
            **kwargs: Optional keyword arguments to Slisemap.lbfgs.

        Returns:
            float: The loss value.
        """
        loss = np.repeat(np.inf, 2)
        loss[0] = self.lbfgs(max_iter=max_iter, only_B=True, **kwargs)
        if verbose:
            print(f"LBFGS   0: {loss[0]:.2f}")
        cc = CheckConvergence(patience)
        for i in range(max_escapes):
            self.escape(escape_fn=escape_fn)
            loss[1] = self.value()
            if verbose:
                print(f"Escape {i:2d}: {loss[1]:.2f}")
            if noise > 0.0:
                self._Z = torch.normal(self.Z, noise, generator=self._random_state)
            loss[0] = self.lbfgs(max_iter=max_iter, **kwargs)
            if verbose:
                print(f"LBFGS  {i+1:2d}: {loss[0]:.2f}")
            if cc.has_converged(loss, self.copy):
                break
        if cc.optimal_state is not None:
            self._Z = cc.optimal_state._Z
            self._B = cc.optimal_state._B
        return self.lbfgs(max_iter=max_iter * 2, **kwargs)

    optimize = optimise

    def fit_new(
        self,
        Xnew: Union[np.ndarray, torch.Tensor],
        ynew: Union[float, np.ndarray, torch.Tensor],
        optimise: bool = True,
        between: bool = True,
        escape_fn: Callable = escape_neighbourhood,
        loss: bool = False,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]:
        """Generate embedding(s) and model(s) for new data item(s).

        This works as follows:
            1. Find good initial embedding(s) and local model(s) using the escape_fn.
            2. Optionally finetune the embedding(s) and model(s) using LBFG.

        Args:
            Xnew (Union[np.ndarray, torch.Tensor]): New data point(s).
            ynew (Union[float, np.ndarray, torch.Tensor]): New target(s).
            optimise (bool, optional): Should the embedding and model be optimised (after finding the neighbourhood). Defaults to True.
            between (bool, optional): If `optimise=True` should the new points affect each other. Defaults to True.
            escape_fn (Callable, optional): Escape function to use as initialisation (escape_neighbourhood/escape_greedy/escape_marginal). Defaults to escape_neighbourhood.
            loss (bool, optional): Return a vector of individual losses for the new items. Defaults to False.
            **kwargs: Optional keyword arguments to LBFGS.

        Returns:
            Tuple[np.ndarray, np.ndarray, Optional[np.ndarray]]: Bnew, Znew, (loss).
        """
        Xnew = self._as_new_X(Xnew)
        n = Xnew.shape[0]
        ynew = self._as_new_Y(ynew, n)
        Bnew, Znew = escape_fn(
            X=Xnew,
            Y=ynew,
            B=self.B,
            Z=self.Z,
            local_model=self.local_model,
            local_loss=self.local_loss,
            distance=self.distance,
            kernel=self.kernel,
            radius=self.radius,
            force_move=False,
            Xold=self.X,
            Yold=self.Y,
            jit=self.jit,
        )

        if optimise:
            lf, set_new = make_marginal_loss(
                X=self.X,
                Y=self.Y,
                B=self.B,
                Z=self.Z,
                Xnew=Xnew if between else Xnew[:1],
                Ynew=ynew if between else ynew[:1],
                local_model=self.local_model,
                local_loss=self.local_loss,
                distance=self.distance,
                kernel=self.kernel,
                radius=self.radius,
                lasso=self.lasso,
                ridge=self.ridge,
                jit=self.jit,
            )
            if between:
                Bnew = Bnew.detach().requires_grad_(True)
                Znew = Znew.detach().requires_grad_(True)
                LBFGS(lambda: lf(Bnew, Znew), [Bnew, Znew], **kwargs)
            else:
                for j in range(n):
                    set_new(Xnew[None, j], ynew[None, j])
                    Bi = Bnew[None, j].detach().clone().requires_grad_(True)
                    Zi = Znew[None, j].detach().clone().requires_grad_(True)
                    LBFGS(lambda: lf(Bi, Zi), [Bi, Zi], **kwargs)
                    Bnew[j] = Bi.detach()
                    Znew[j] = Zi.detach()
        if self.radius > 0:
            Zout = Znew * (self.radius / (torch.sum(self._Z**2) + 1e-8))
        else:
            Zout = Znew
        if loss:
            lf = self.get_loss_fn(individual=True)
            if between:
                loss = lf(
                    X=torch.cat((self.X, Xnew), 0),
                    Y=torch.cat((self.Y, ynew), 0),
                    B=torch.cat((self.B, Bnew), 0),
                    Z=torch.cat((self.Z, Znew), 0),
                )[self.n :]
                loss = _tonp(loss)
            else:
                if self._jit:
                    lf = torch.jit.trace(lf, (Xnew[:1], ynew[:1], Bnew[:1], Znew[:1]))
                loss = np.zeros(n)
                for j in range(n):
                    l = lf(
                        X=torch.cat((self.X, Xnew[None, j]), 0),
                        Y=torch.cat((self.Y, ynew[None, j]), 0),
                        B=torch.cat((self.B, Bnew[None, j]), 0),
                        Z=torch.cat((self.Z, Znew[None, j]), 0),
                    )[-1]
                    loss[j] = l.detach().cpu().item()
            return _tonp(Bnew), _tonp(Zout), loss
        else:
            return _tonp(Bnew), _tonp(Zout)

    def predict(
        self,
        Xnew: Union[np.ndarray, torch.Tensor],
        Znew: Union[np.ndarray, torch.Tensor],
        **kwargs,
    ) -> np.ndarray:
        """Predict new outcomes when the data and embedding is known.

        Args:
            Xnew (Union[np.ndarray, torch.Tensor]): Data matrix.
            Znew (Union[np.ndarray, torch.Tensor]): Embedding matrix.
            **kwargs: Optional keyword arguments to LBFGS.

        Returns:
            np.ndarray: Prediction matrix.
        """
        Xnew = torch.as_tensor(Xnew, **self.tensorargs)
        Xnew = torch.atleast_2d(Xnew)
        N = Xnew.shape[0]
        if self._intercept:
            Xnew = torch.cat([Xnew, torch.ones([N, 1], **self.tensorargs)], 1)
        _assert(
            Xnew.shape == (N, self.m),
            f"Xnew has the wrong shape {Xnew.shape} != {(N, self.m)}",
        )
        Znew = torch.as_tensor(Znew, **self.tensorargs)
        Znew = torch.atleast_2d(Znew)
        _assert(
            Znew.shape == (N, self.d),
            f"Znew has the wrong shape {Znew.shape} != {(N, self.d)}",
        )
        D = self._distance(Znew, self._Z)
        W = self.kernel(D)
        Bnew = self._B[torch.argmin(D, 1)].clone().requires_grad_(True)
        loss = lambda: (
            torch.sum(
                W * self.local_loss(self.local_model(self._X, Bnew), self._Y, Bnew)
            )
            + self.lasso * torch.sum(torch.abs(Bnew))
            + self.ridge * torch.sum(Bnew**2)
        )
        LBFGS(loss, [Bnew], **kwargs)
        return _tonp(torch.diagonal(self.local_model(Xnew, Bnew), dim1=0, dim2=1).T)

    def copy(self) -> "Slisemap":
        """Make a copy of this Slisemap that references as much of the same torch-data as possible.

        Returns:
            Slisemap: An almost shallow copy of this Slisemap object.
        """
        other = copy(self)  # Shallow copy!
        # Deep copy these:
        other._B = other._B.clone().detach()
        other._Z = other._Z.clone().detach()
        return other

    def save(self, f: Union[str, PathLike, BinaryIO], **kwargs):
        """Save the Slisemap object to a file.
        This method uses `torch.save`, which uses `pickle` for the non-pytorch properties.
        This comes with the normal caveats like lambda-functions not being supported.
        However, the default pickle module can be overridden (see `torch.save`).
        The default ending for `torch.save` is ".pt" and by default the file is compressed.

        Args:
            f (Union[str, PathLike, BinaryIO]): Either a Path-like object or a (writable) File-like object.
            **kwargs: Parameters forwarded to `torch.save`.
        """
        loss = self._loss
        try:
            self._B = self._B.detach()
            self._Z = self._Z.detach()
            self._loss = None
            torch.save(self, f, **kwargs)
        finally:
            self._loss = loss

    @classmethod
    def load(
        cls,
        f: Union[str, PathLike, BinaryIO],
        device: Union[None, str, torch.device] = None,
        **kwargs,
    ) -> "Slisemap":
        """Load a Slisemap object from a file.
        This method uses `torch.load`, which uses `pickle` for the non-pytorch properties.
        However, the default pickle module can be overridden (see `torch.load`).

        Args:
            f (Union[str, PathLike, BinaryIO]): Either a Path-like object or a (readable) File-like object.
            device (Union[None, str, torch.device], optional): Device to load the tensors to (or the original if None). Defaults to None.
            **kwargs: Parameters forwarded to `torch.load`.

        Returns:
            Slisemap: The loaded Slisemap object.
        """
        sm = torch.load(f, map_location=device, **kwargs)
        return sm

    def _cluster_models(
        self, clusters: int, B: Optional[np.ndarray] = None, random_state: int = 42
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Get cluster labels from k-means on the local model coefficients.
        This function uses KMeans from scikit-learn.

        Args:
            clusters (int): Number of clusters.
            B (Optional[np.ndarray], optional): B matrix. Defaults to self.get_B().
            random_state (int, optional): random_state for the KMeans clustering. Defaults to 42.

        Returns:
            Tuple[np.ndarray, np.ndarray]: vector of cluster labels an matrix of cluster centers.
        """
        from sklearn.cluster import KMeans

        km = KMeans(clusters, random_state=random_state).fit(
            B if B is not None else self.get_B()
        )
        influence = np.abs(km.cluster_centers_)
        influence = influence.max(0) + influence.mean(0)
        col = np.argmax(influence)
        ord = np.argsort(km.cluster_centers_[:, col])
        return np.argsort(ord)[km.labels_], km.cluster_centers_[ord]

    def plot(
        self,
        title: str = "",
        variables: Optional[Sequence[str]] = None,
        targets: Union[None, str, Sequence[str]] = None,
        clusters: Union[None, int, np.ndarray] = None,
        bars: Union[bool, int] = False,
        jitter: Union[float, np.ndarray] = 0.0,
        B: Optional[np.ndarray] = None,
        Z: Optional[np.ndarray] = None,
        show: bool = True,
        **kwargs,
    ) -> Optional[Figure]:
        """Plot the Slisemap solution using seaborn.

        Args:
            title (str, optional): Title of the plot. Defaults to "".
            variables (Optional[Sequence[str]], optional): List of variable names. Defaults to None.
            targets (Union[None, str, Sequence[str]], optional): Target name(s). Defaults to None.
            clusters (Union[None, int, np.ndarray], optional): Can be None (plot individual losses), an int (plot k-means clusters of B), or an array of known cluster id:s. Defaults to None.
            bars (Union[bool, int], optional): If the clusters are from k-means, plot the local models in a bar plot. If `bar` is an int then only plot the most influential variables. Defaults to False.
            jitter (Union[float, np.ndarray], optional): Add random (normal) noise to the embedding, or a matrix with pre-generated noise matching Z. Defaults to 0.0.
            B (Optional[np.ndarray], optional): Override self.get_B() in the plot. Defaults to None.
            Z (Optional[np.ndarray], optional): Override self.get_Z() in the plot. Defaults to None.
            show (bool, optional): Show the plot. Defaults to True.
            **kwargs: Additional arguments to `plt.subplots`.

        Returns:
            Optional[Figure]: Matplotlib figure if show=False.
        """
        Z = self.get_Z(rotate=True) if Z is None else Z
        B = self.get_B() if B is None else B
        if Z.shape[1] == 1:
            Z = np.concatenate((Z, np.zeros_like(Z)), 1)
        elif Z.shape[1] > 2:
            _warn(
                "Only the first two dimensions in the embedding are plotted",
                Slisemap.plot,
            )
        if not isinstance(jitter, float):
            Z = Z + jitter
        elif jitter > 0:
            Z = np.random.normal(Z, jitter)
        kwargs.setdefault("figsize", (12, 6))
        fig, (ax1, ax2) = plt.subplots(1, 2, **kwargs)
        if variables is not None:
            if self._intercept and len(variables) == self.m - 1:
                variables = list(variables) + ["Intercept"]
            if targets and not isinstance(targets, str) and len(targets) > 0:
                if B.shape[1] % len(variables) == 0 and B.shape[1] % len(targets) == 0:
                    variables = [f"{t}: {v}" for t in targets for v in variables]
                    variables = variables[: B.shape[1]]
            _assert(
                len(variables) == B.shape[1],
                f"The number of variable names ({len(variables)}) must match the number of coefficients ({B.shape[1]})",
            )
        if isinstance(clusters, int):
            clusters, centers = self._cluster_models(clusters, B)
            if bars:
                y = np.arange(B.shape[1]) if variables is None else variables
                if not isinstance(bars, bool):
                    influence = np.abs(centers)
                    influence = influence.max(0) + influence.mean(0)
                    mask = np.argsort(-influence)[:bars]
                    y = np.asarray(y)[mask]
                    B = B[:, mask]
                g = sns.barplot(
                    y=np.tile(y, B.shape[0]),
                    x=B.ravel(),
                    hue=np.repeat(clusters, B.shape[1]),
                    ax=ax2,
                    palette="bright",
                    orient="h",
                )
                g.legend().remove()
                lim = np.max(np.abs(g.get_xlim()))
                g.set(xlabel=None, ylabel=None, xlim=(-lim, lim))
            else:
                B = centers
        else:
            B = B[np.argsort(Z[:, 0])]
            _assert(not bars, "`bar!=False` requires that `clusters` is an integer")
        if clusters is None:
            L = self.value(individual=True)
            sns.scatterplot(x=Z[:, 0], y=Z[:, 1], hue=L, palette="crest", ax=ax1)
            sns.heatmap(B, ax=ax2, center=0, cmap="RdBu", robust=True)
            ax1.legend(title="Fidelity")
            if variables is not None:
                ax2.set_xticks(np.arange(len(variables)) + 0.5)
                ax2.set_xticklabels(variables, rotation=30)
        else:
            sns.scatterplot(
                x=Z[:, 0],
                y=Z[:, 1],
                hue=clusters,
                style=clusters,
                palette="bright",
                ax=ax1,
            )
            ax1.legend(title="Cluster")
            if not bars:
                sns.heatmap(B, ax=ax2, center=0, cmap="RdBu", robust=True)
                if variables is not None:
                    ax2.set_xticks(np.arange(len(variables)) + 0.5)
                    ax2.set_xticklabels(variables, rotation=30)
        ax1.set_xlabel("SLISEMAP 1")
        ax1.set_ylabel("SLISEMAP 2")
        ax1.axis("equal")
        ax2.set_xlabel("Coefficients")
        ax1.set_title("Embedding")
        ax2.set_title("Local Models")
        sns.despine(fig)
        plt.suptitle(title)
        plt.tight_layout()
        if show:
            plt.show()
        else:
            return fig

    def plot_position(
        self,
        X: Union[None, np.ndarray, torch.Tensor] = None,
        Y: Union[None, float, np.ndarray, torch.Tensor] = None,
        index: Union[None, int, Sequence[int]] = None,
        title: str = "",
        jitter: Union[float, np.ndarray] = 0.0,
        col_wrap: int = 4,
        selection: bool = True,
        legend_inside: bool = True,
        Z: Optional[np.ndarray] = None,
        show: bool = True,
        **kwargs,
    ) -> Optional[sns.FacetGrid]:
        """Plot fidelities for alternative locations for the selected item(s).
        Indicate the selected item(s) either via X&Y or via index.

        Args:
            X (Union[None, np.ndarray, torch.Tensor], optional): Data matrix for the selected data item(s). Defaults to None.
            Y (Union[None, float, np.ndarray, torch.Tensor], optional): Response matrix for the selected data item(s). Defaults to None.
            index (Union[None, int, Sequence[int]], optional): Index/indices of the selected data item(s). Defaults to None.
            title (str, optional): Title of the plot. Defaults to "".
            jitter (Union[float, np.ndarray], optional): Add random (normal) noise to the embedding, or a matrix with pre-generated noise matching Z. Defaults to 0.0.
            col_wrap (int, optional): Maximum number of columns. Defaults to 4.
            selection (bool, optional): Mark the selected data item(s), if index is given. Defaults to True.
            legend_inside (bool, optional): Move the legend inside the grid (if there is an empty cell). Defaults to True.
            Z (Optional[np.ndarray], optional): Override self.get_Z() in the plot. Defaults to None.
            show (bool, optional): Show the plot. Defaults to True.
            **kwargs: Additional arguments to seaborn.relplot.

        Returns:
            Optional[sns.FacetGrid]: Seaborn FacetGrid if show=False.
        """
        import pandas as pd

        Z = Z if Z is not None else self.get_Z(rotate=True)
        if Z.shape[1] == 1:
            Z = np.concatenate((Z, np.zeros_like(Z)), 1)
        elif Z.shape[1] > 2:
            _warn(
                "Only the first two dimensions in the embedding are plotted",
                Slisemap.plot_position,
            )
        if not isinstance(jitter, float):
            Z = Z + jitter
        elif jitter > 0:
            Z = np.random.normal(Z, jitter)
        if index is None:
            _assert(
                X is not None and Y is not None, "Either index or X and Y must be given"
            )
            L = self.get_L(X, Y)
        else:
            if isinstance(index, int):
                index = [index]
            L = self.get_L()[:, index]
        df = pd.concat(
            [
                pd.DataFrame(
                    {
                        "SLISEMAP 1": Z[:, 0],
                        "SLISEMAP 2": Z[:, 1],
                        "Fidelity": loss,
                        "i": i,
                    }
                )
                for i, loss in enumerate(L.T)
            ],
            ignore_index=True,
        )
        kwargs.setdefault("palette", "crest")
        kwargs.setdefault("kind", "scatter")
        g: sns.FacetGrid = sns.relplot(
            data=df,
            x="SLISEMAP 1",
            y="SLISEMAP 2",
            hue="Fidelity",
            hue_norm=tuple(np.quantile(df["Fidelity"], (0.0, 0.9))),
            col="i",
            col_wrap=min(col_wrap, L.shape[1]),
            legend=False,
            **kwargs,
        )
        cmap = plt.get_cmap(kwargs["palette"])
        legend = {
            f"{l:.{2}f}": Patch(color=cmap(n))
            for l, n in zip(
                np.linspace(*np.quantile(df["Fidelity"], (0.0, 0.9)), 6).round(2),
                np.linspace(0.0, 1.0, 6),
            )
        }
        inside = legend_inside and col_wrap < L.shape[1] and L.shape[1] % col_wrap != 0
        w = 1 / col_wrap
        h = 1 / ((L.shape[1] - 1) // col_wrap + 1)
        if selection and index is not None:
            for i, ax in zip(index, g.axes.ravel()):
                ax.scatter(Z[i, 0], Z[i, 1], 100, "#9b52b4", "X")
            g.add_legend(
                legend,
                "Fidelity",
                loc="lower center" if inside else "upper right",
                bbox_to_anchor=(1 - w, h * 0.4, w * 0.9, h * 0.5) if inside else None,
            )
            g.add_legend(
                {"": Line2D([], [], None, "None", "#9b52b4", "X", 8)},
                "Selected",
                loc="upper center" if inside else "lower right",
                bbox_to_anchor=(1 - w, 0.0, w * 0.9, h * 0.4) if inside else None,
            )
        else:
            g.add_legend(
                legend,
                "Fidelity",
                loc="center" if inside else "center right",
                bbox_to_anchor=(1 - w, 0, w * 0.9, h) if inside else None,
            )
        g.set_titles("")
        plt.suptitle(title)
        if inside:
            plt.tight_layout()
        else:
            g.tight_layout()
        if show:
            plt.show()
        else:
            return g

    def plot_dist(
        self,
        title: str = "",
        X: Optional[np.ndarray] = None,
        Y: Optional[np.ndarray] = None,
        variables: Optional[List[str]] = None,
        targets: Union[None, str, Sequence[str]] = None,
        clusters: Union[None, int, np.ndarray] = None,
        scatter: bool = False,
        jitter: float = 0.0,
        col_wrap: int = 4,
        legend_inside: bool = True,
        show: bool = True,
        **kwargs,
    ) -> Optional[sns.FacetGrid]:
        """Plot the distribution of the variables, either as density plots (with clusters) or as scatterplots.

        Args:
            title (str, optional): Title of the plot. Defaults to "".
            X (Optional[np.ndarray], optional): Replacement data matrix (e.g. without normalisation). Defaults to None.
            Y (Optional[np.ndarray], optional): Replacement target matrix (e.g. without normalisation). Defaults to None.
            variables (Optional[List[str]], optional): List of variable names. Defaults to None.
            targets (Union[None, str, Sequence[str]], optional): Target name(s). Defaults to None.
            clusters (Union[None, int, np.ndarray], optional): Number of cluster or vector of cluster labels. Defaults to None.
            scatter (bool, optional): Use scatterplots instead of density plots (clusters are ignored). Defaults to False.
            jitter (float, optional): Add jitter to the scatterplots. Defaults to 0.0.
            col_wrap (int, optional): Maximum number of columns. Defaults to 4.
            legend_inside (bool, optional): Move the legend inside the grid (if there is an empty cell). Defaults to True.
            show (bool, optional): Show the plot. Defaults to True.
            **kwargs: Additional arguments to seaborn.relplot.

        Returns:
            Optional[sns.FacetGrid]: Seaborn FacetGrid if show=False.
        """
        import pandas as pd

        if X is None:
            X = self.get_X(intercept=False)
        if Y is None:
            Y = self.get_Y()
        else:
            Y = np.reshape(Y, (X.shape[0], -1))
        if variables is None:
            variables = [f"Variable {i}" for i in range(self.m - self._intercept)]
        if targets is None:
            targets = (
                [f"Target {i}" for i in range(self.p)] if self.p > 1 else ["Target"]
            )
        elif isinstance(targets, str):
            targets = [targets]

        if not scatter:
            if isinstance(clusters, int):
                clusters, _ = self._cluster_models(clusters)
            elif clusters is None:
                legend_inside = False
            df = pd.concat(
                [
                    pd.DataFrame(dict(var=n, Value=XY[:, i], Cluster=clusters))
                    for v, XY in [(targets, Y), (variables, X)]
                    for i, n in enumerate(v)
                ],
                ignore_index=True,
            )
            if kwargs.setdefault("kind", "kde") == "kde":
                kwargs.setdefault("bw_adjust", 0.75)
                kwargs.setdefault("common_norm", False)
            kwargs.setdefault("palette", "bright")
            kwargs.setdefault("facet_kws", dict(sharex=False, sharey=False))
            g: sns.FacetGrid = sns.displot(
                data=df,
                x="Value",
                hue=None if clusters is None else "Cluster",
                col="var",
                col_wrap=col_wrap,
                **kwargs,
            )
            g.set_titles("{col_name}")
            g.set_xlabels("")
        else:
            Z = self.get_Z(rotate=True)
            if Z.shape[1] == 1:
                Z = np.concatenate((Z, np.zeros_like(Z)), 1)
            elif Z.shape[1] > 2:
                _warn(
                    "Only the first two dimensions in the embedding are plotted",
                    Slisemap.plot_dist,
                )
            if jitter > 0:
                Z = np.random.normal(Z, jitter)
            df = pd.concat(
                [
                    pd.DataFrame(
                        {
                            "var": n,
                            "Value": XY[:, i],
                            "SLISEMAP 1": Z[:, 0],
                            "SLISEMAP 2": Z[:, 1],
                        }
                    )
                    for v, XY in [(targets, Y), (variables, X)]
                    for i, n in enumerate(v)
                ],
                ignore_index=True,
            )
            kwargs.setdefault("palette", "rocket")
            kwargs.setdefault("kind", "scatter")
            g = sns.relplot(
                data=df,
                x="SLISEMAP 1",
                y="SLISEMAP 2",
                hue="Value",
                col="var",
                col_wrap=col_wrap,
                **kwargs,
            )
            g.set_titles("{col_name}")
        plt.suptitle(title)
        cells = len(targets) + len(variables)
        if legend_inside and col_wrap < cells and cells % col_wrap != 0:
            w = 1 / col_wrap
            h = 1 / ((cells - 1) // col_wrap + 1)
            sns.move_legend(
                g,
                "center",
                bbox_to_anchor=(1 - w, h * 0.1, w * 0.9, h * 0.9),
                frameon=False,
            )
            plt.tight_layout()
        else:
            g.tight_layout()
        if show:
            plt.show()
        else:
            return g



"""
This module contains the SLISEMAP loss functions.
"""

from typing import Callable

import torch
from torch.nn.functional import softmax

from slisemap.utils import _warn


def softmax_kernel(D: torch.Tensor) -> torch.Tensor:
    """Kernel function based on softmax.

    Args:
        D (torch.Tensor): Distance matrix.

    Returns:
        torch.Tensor: Weight matrix.
    """
    return softmax(-D, 1)


def make_loss(
    local_model: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    local_loss: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
    distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    kernel: Callable[[torch.Tensor], torch.Tensor],
    radius: float = 3.5,
    lasso: float = 0.0,
    ridge: float = 0.0,
    z_norm: float = 1.0,
    individual: bool = False,
) -> Callable[[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor,], torch.Tensor,]:
    """Create a loss function for Slisemap to optimise

    Args:
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Prediction function for the local models.
        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]): Loss function for the local models.
        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Embedding distance function.
        kernel (Callable[[torch.Tensor], torch.Tensor]): Kernel for embedding distances.
        radius (float, optional): For enforcing the radius of Z. Defaults to 3.5.
        lasso (float, optional): Lasso-regularisation coefficient for B ($\\lambda_{lasso} * ||B||_1$). Defaults to 0.0.
        ridge (float, optional): Ridge-regularisation coefficient for B ($\\lambda_{ridge} * ||B||_2$). Defaults to 0.0.
        z_norm (float, optional): Z normalisation regularisation coefficient ($\\lambda_{norm} * (sum(Z^2)-n)^2$). Defaults to 1.0.
        individual (bool, optional): Return individual (row-wise) losses. Defaults to False.

    Returns:
        Callable[[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]: Loss function for SLISEMAP
    """
    dim = 1 if individual else ()
    if individual and z_norm > 0:
        _warn(
            "The Z normalisation is added to every individual loss if z_norm > 0",
            make_loss,
        )

    def loss(
        X: torch.Tensor,
        Y: torch.Tensor,
        B: torch.Tensor,
        Z: torch.Tensor,
    ) -> torch.Tensor:
        """Slisemap loss function.

        Args:
            X (torch.Tensor): Data matrix.
            Y (torch.Tensor): Target matrix.
            B (torch.Tensor): Local models.
            Z (torch.Tensor): Embedding matrix [n, d].

        Returns:
            torch.Tensor: loss value.
        """
        if radius > 0:
            Zss = torch.sum(Z**2)
            Z = Z * (radius / (torch.sqrt(Zss / Z.shape[0]) + 1e-8))
        D = distance(Z, Z)
        L = local_loss(local_model(X, B), Y, B)
        a = torch.sum(kernel(D) * L, dim=dim)
        if lasso > 0:
            a += lasso * torch.sum(B.abs(), dim=dim)
        if ridge > 0:
            a += ridge * torch.sum(B**2, dim=dim)
        if z_norm > 0 and radius > 0:
            a += z_norm * (Zss - Z.shape[0]) ** 2
        return a

    return loss


def make_marginal_loss(
    X: torch.Tensor,
    Y: torch.Tensor,
    B: torch.Tensor,
    Z: torch.Tensor,
    Xnew: torch.Tensor,
    Ynew: torch.Tensor,
    local_model: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    local_loss: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
    distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    kernel: Callable[[torch.Tensor], torch.Tensor],
    radius: float = 3.5,
    lasso: float = 0.0,
    ridge: float = 0.0,
    jit: bool = True,
) -> Callable[[torch.Tensor, torch.Tensor], torch.Tensor]:
    """Create a loss for adding new points with Slisemap.

    Args:
        X (torch.Tensor): The existing data matrix [n_old, m].
        Y (torch.Tensor): The existing target matrix [n_old, k].
        B (torch.Tensor): The fitted models [n_old, p].
        Z (torch.Tensor): The fitted embedding [n_old, d].
        Xnew (torch.Tensor): The new data matrix [n_new, m].
        Ynew (torch.Tensor): The new target matrix [n_new, k].
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Prediction function for the local models.
        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]): Loss function for the local models.
        kernel (Callable[[torch.Tensor], torch.Tensor]): Kernel for embedding distances.
        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Embedding distance function.
        radius (float, optional): For enforcing the radius of Z. Defaults to 3.5.
        lasso (float, optional): Lasso-regularisation coefficient for B ($\\lambda_{lasso} * ||B||_1$). Defaults to 0.0.
        ridge (float, optional): Ridge-regularisation coefficient for B ($\\lambda_{ridge} * ||B||_2$). Defaults to 0.0.
        jit (bool, optional): Just-In-Time compile the loss function. Defaults to True.

    Returns:
        Callable[[torch.Tensor, torch.Tensor], torch.Tensor]: Loss function that takes `Bnew` [n_new, p] and `Znew` [n_new, d] (or `Xnew`, `Ynew`, `Bnew` and `Znew` if `Xnew` and `Ynew` were not give above).
    """
    Xcomb = torch.cat((X, Xnew), 0)
    Ycomb = torch.cat((Y, Ynew), 0)
    Nold = X.shape[0]
    L0 = local_loss(local_model(Xcomb, B), Ycomb, B)  # Nold x Ncomb
    D0 = distance(Z, Z)  # Nold x Nold

    def set_new(Xnew: torch.Tensor, Ynew: torch.Tensor):
        nonlocal Xcomb, Ycomb, L0
        Xcomb[Nold:] = Xnew
        Ycomb[Nold:] = Ynew
        L0[:, Nold:] = local_loss(local_model(Xnew, B), Ynew, B)

    if radius > 0:
        Zss0 = torch.sum(Z**2)

    def loss(Bnew: torch.Tensor, Znew: torch.Tensor) -> torch.Tensor:
        L1 = local_loss(local_model(Xcomb, Bnew), Ycomb, Bnew)  # Nnew x Ncomb
        L = torch.cat((L0, L1), 0)  # Ncomb x Ncomb

        D1 = distance(Znew, Z)  # Nnew x Nold
        D2 = distance(Znew, Znew)  # Nnew x Nnew
        D3 = D1.transpose(0, 1)
        D = torch.cat(
            (torch.cat((D0, D1), 0), torch.cat((D3, D2), 0)), 1
        )  # Ncomb x Ncomb
        if radius > 0:
            Zss = Zss0 + torch.sum(Znew**2)
            Ncomb = Z.shape[0] + Znew.shape[0]
            norm = radius / (torch.sqrt(Zss / Ncomb) + 1e-8)
            D = D * norm

        kD = kernel(D)
        a = torch.sum(kD * L)
        if lasso > 0:
            a += lasso * torch.sum(Bnew.abs())
        if ridge > 0:
            a += ridge * torch.sum(Bnew**2)
        return a

    if jit:
        Nnew = Xnew.shape[0]
        loss = torch.jit.trace(loss, (B[:1].expand(Nnew, -1), Z[:1].expand(Nnew, -1)))
    return loss, set_new



"""
SLISEMAP: Combine local explanations with supervised dimensionality reduction
===============================================================================

SLISEMAP is a supervised dimensionality reduction method, that takes data, in the form of
vectors, and predictions from a "black box" regression or classification model as input.
SLISEMAP then simultaneously finds local explanations for all data items and builds a
(typically) two-dimensional global visualisation of the black box model such that data
items with similar local explanations are projected nearby. The explanations consists of
"white box" models that locally approximate the "black box" model.

SLISEMAP uses *PyTorch* for efficient optimisation, and optional GPU-acceleration. For
more information see the the `repository <https://github.com/edahelsinki/slisemap>`_ or
the `paper <https://arxiv.org/abs/2201.04455>`_.

Citation
--------
| Björklund, A., Mäkelä, J. & Puolamäki, K. (2022).  
| SLISEMAP: Supervised dimensionality reduction through local explanations.  
| arXiv:2201.04455 [cs], https://arxiv.org/abs/2201.04455.  


Example Usage
-------------
.. code-block:: python

    from slisemap import Slisemap
    import numpy as np

    X = np.array([[0.1,0.5,0.7], [0.8,0.9,1], [0.8,0.5,0.3], [0.1,0.2,0.3], [1,2,5], [2,3,4], [2,0,1]])
    y = np.array([1, 2, 3, 4, 1.5, 1.8, 1.7])
    sm = Slisemap(X, y, radius=3.5, lasso=1e-4, ridge=2e-4)
    sm.optimise()
    sm.plot()
"""

from slisemap.slisemap import Slisemap



"""
These are diagnostics for identifying potential issues with Slisemap solutions.

Typical usage:

.. code-block:: python

        sm = Slisemap(...)
        sm.optimise()
        diagnostics = diagnose(sm)
        print_diagnostics(diagnostics)
        plot_diagnostics(sm, diagnostics)
"""
from functools import reduce
from typing import Dict, Optional, Tuple, Union

import numpy as np
import seaborn as sns
import torch
from matplotlib import pyplot as plt

from slisemap.slisemap import Slisemap
from slisemap.utils import _tonp, global_model


def _size(n: int, part: Union[float, int]) -> int:
    if isinstance(part, int):
        return part
    else:
        return int(n * part)


def _frac(n: int, part: Union[float, int]) -> float:
    if isinstance(part, int):
        return part / n
    else:
        return part


def global_model_losses(
    sm: Slisemap, indices: Optional[np.ndarray] = None, **kwargs
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Train a global model

    Args:
        sm (Slisemap): Slisemap object.
        indices (Optional[np.ndarray], optional): Optional subsampling indices. Defaults to None.
        **kwargs: Optional keyword arguments to LBFGS.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Tuple with vector of individual losses and vector of model coefficients.
    """
    if indices is None:
        X = sm.X
        Y = sm.Y
    else:
        X = sm.X[indices]
        Y = sm.Y[indices]
    B = global_model(
        X=X,
        Y=Y,
        local_model=sm.local_model,
        local_loss=sm.local_loss,
        coefficients=sm.coefficients,
        lasso=sm.lasso,
        ridge=sm.ridge,
    )
    return sm.local_loss(sm.local_model(sm.X, B), sm.Y).detach()


def print_diagnostics(diagnostics: Dict[str, np.ndarray], summary: bool = False):
    """Print diagnostic results.

    Args:
        diagnostics (Dict[str, np.ndarray]): Dictionary of diagnostic results.
        summary (bool, optional): Print only one summary for all the diagnostics. Defaults to False.
    """
    if summary:
        issues = reduce(lambda a, b: a + b.astype(int), diagnostics.values())
        if np.sum(issues) == 0:
            print(f"All data items passed all the diagnostics!")
        else:
            print(
                f"{np.mean(issues > 0) * 100 :.1f}% of the data items failed at least",
                "one diagnostic and the average failure rate across all diagnostics",
                f"was {np.sum(issues) / (len(issues) * len(diagnostics)) *100:.1f}%",
            )
    else:
        for name, mask in diagnostics.items():
            rate = np.mean(mask) * 100
            if rate == 0:
                print(f"All data items passed the `{name}` diagnostic!")
            else:
                print(f"{rate:.1f}% of the data items failed the `{name}` diagnostic!")


def plot_diagnostics(
    Z: Union[Slisemap, np.ndarray],
    diagnostics: Dict[str, np.ndarray],
    summary: bool = False,
    **kwargs,
) -> sns.FacetGrid:
    """Plot diagnostic results (remember to call `plt.show`).

    Args:
        Z (Union[Slisemap, np.ndarray]): The Slisemap object, or embedding matrix.
        diagnostics (Dict[str, np.ndarray]): Dictionary of diagnostic results.
        summary (bool, optional): Combine multiple diagnostics into one plot. Defaults to False.
        **kwargs: Additional parameters to `seaborn.relplot`.

    Returns:
        seaborn.FacetGrid: The plot as a `seaborn.FacetGrid`.
    """
    import pandas as pd

    if isinstance(Z, Slisemap):
        Z = Z.get_Z(rotate=True)
    if len(diagnostics) == 1:
        for name, mask in diagnostics.items():
            df = pd.DataFrame({"Z1": Z[:, 0], "Z2": Z[:, 1], name: mask})
            g = sns.relplot(
                data=df, x="Z1", y="Z2", hue=name, style=name, kind="scatter", **kwargs
            )
    elif summary:
        issues = reduce(lambda a, b: a + b.astype(int), diagnostics.values())
        df = pd.DataFrame(
            {"Z1": Z[:, 0], "Z2": Z[:, 1], "Problem": issues > 0, "Severity": issues}
        )
        g = sns.relplot(
            data=df,
            x="Z1",
            y="Z2",
            hue="Severity",
            style="Problem",
            kind="scatter",
            **kwargs,
        )
    else:
        if "col_wrap" not in kwargs:
            if len(diagnostics) <= 4:
                kwargs["col_wrap"] = len(diagnostics)
            elif len(diagnostics) < 7 or len(diagnostics) == 9:
                kwargs["col_wrap"] = 3
            else:
                kwargs["col_wrap"] = 4
        df = pd.concat(
            [
                pd.DataFrame(
                    {
                        "Z1": Z[:, 0],
                        "Z2": Z[:, 1],
                        "Problem": mask,
                        "Diagnostic": f"{diag} ({np.mean(mask) * 100:.1f} %)",
                    }
                )
                for diag, mask in diagnostics.items()
            ]
        )
        g = sns.relplot(
            data=df,
            x="Z1",
            y="Z2",
            hue="Problem",
            style="Problem",
            kind="scatter",
            col="Diagnostic",
            **kwargs,
        )
    plt.suptitle("Slisemap Diagnostics")
    return g


def distant_diagnostic(sm: Slisemap, max_distance: float = 10.0) -> np.ndarray:
    """Check if any data item in the embedding is too far way.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        max_distance (float, optional): Maximum distance from origo in the embedding. Defaults to 10.0.

    Returns:
        np.ndarray: Boolean mask of problematic data items.
    """
    return np.sum(sm.get_Z() ** 2, 1) > max_distance**2


def heavyweight_diagnostic(
    sm: Slisemap, min_size: Union[float, int] = 0.1
) -> np.ndarray:
    """Check if any data item has a self-weight that is too large.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        min_size (Union[float, int], optional): Miniumum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.1.

    Returns:
        np.ndarray: Boolean mask of problematic data items.
    """
    return _tonp(sm.get_W(numpy=False).diag() > _frac(sm.n, min_size))


def lightweight_diagnostic(
    sm: Slisemap, max_size: Union[float, int] = 0.5
) -> np.ndarray:
    """Check if any data item has a self-weight that is too small

    Args:
        sm (Slisemap): Trained Slisemap solution.
        max_size (Union[float, int], optional): Maximum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.5.

    Returns:
        np.ndarray: Boolean mask of problematic data items.
    """
    return _tonp(sm.get_W(numpy=False).diag() < (1 / _size(sm.n, max_size)))


def weight_neighbourhood_diagnostic(
    sm: Slisemap, min_size: Union[float, int] = 0.1, max_size: Union[float, int] = 0.5
) -> np.ndarray:
    """Check if any data item has a neighbourhood that is too small/large by counting the number of non-lightweight neighbours.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        min_size (Union[float, int], optional): Miniumum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.1.
        max_size (Union[float, int], optional): Maximum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.5.

    Returns:
        np.ndarray: Boolean mask of problematic data items.
    """
    min_size = _size(sm.n, min_size)
    max_size = _size(sm.n, max_size)
    return _tonp((sm.get_W(numpy=False) > (1 / max_size)).sum(1) < min_size)


def loss_neighbourhood_diagnostic(
    sm: Slisemap,
    min_size: Union[float, int] = 0.1,
    smoothing: bool = True,
    median: bool = False,
) -> np.ndarray:
    """Check if any data item has a neighbourhood that is too small/large by comparing local losses to global losses.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        min_size (Union[float, int], optional): Miniumum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.1.
        smoothing (bool, optional): Smooth the sorted losses to avoid sensitivity to outliers. Defaults to True.
        median (bool, optional): Compare against the median global loss instead of the mean global loss. Defaults to False.

    Returns:
        np.ndarray: Boolean mask of problematic data items.
    """
    min_size = _size(sm.n, min_size)
    if median:
        gloss = global_model_losses(sm).median().cpu().item()
    else:
        gloss = global_model_losses(sm).mean().cpu().item()
    llosses = sm.get_L(numpy=False)
    order = torch.argsort(sm.get_D(numpy=False))
    result = np.zeros(sm.n, bool)
    if smoothing:
        filter = torch.as_tensor([[[0.25, 0.5, 0.25]]], **sm.tensorargs)
    for i in range(sm.n):
        lli = llosses[i, order[i]]
        if smoothing:
            lli = torch.conv1d(lli[None, None], filter, padding="same")[0, 0]
        first = torch.where(lli > gloss)[0][:1]
        if first.numel() < 1:
            result[i] = lli.numel() < min_size
        else:
            result[i] = (first < min_size).cpu().item()
    return result


def global_loss_diagnostic(
    sm: Slisemap, bootstrap: int = 10, sd: float = 1.0
) -> np.ndarray:
    """Check if any local model is actually a global model.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        bootstrap (int, optional): Number of (bootstrap) global models to train. Defaults to 10.
        sd (float, optional): Number of standard deviations from the mean (of global models losses) to consider a local model global. Defaults to 1.0.

    Returns:
        np.ndarray: Boolean mask of problematic data items.
    """
    glosses = [
        global_model_losses(sm, np.random.randint(sm.n, size=sm.n)).sum().cpu().item()
        for _ in range(bootstrap)
    ]
    treshold = np.mean(glosses) + np.std(glosses) * sd
    llosses = sm.get_L(numpy=False).sum(1)
    return (llosses < treshold).cpu().numpy()


def optics_diagnostic(sm: Slisemap, min_size: Union[float, int] = 0.1, **kwargs):
    """Use a clustering method to check for problematic data items in the embedding.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        min_size (Union[float, int], optional): Miniumum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.1.

    Returns:
        np.ndarray: Boolean mask of problematic data items.
    """
    from sklearn.cluster import OPTICS

    optics = OPTICS(min_samples=min_size, metric="euclidean", n_jobs=-1, **kwargs)
    return optics.fit(sm.get_Z()).labels_ < 0


def diagnose(
    sm: Slisemap,
    min_size: Union[float, int] = 0.1,
    max_size: Union[float, int] = 0.5,
    max_distance: float = 10.0,
    conservative: bool = False,
) -> Dict[str, np.ndarray]:
    """Run multiple diagnostics.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        min_size (Union[float, int], optional): Miniumum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.1.
        max_size (Union[float, int], optional): Maximum neighbourhood/cluster size (as a fraction or absolute number). Defaults to 0.5.
        max_distance (float, optional): Maximum distance from origo in the embedding. Defaults to 10.0.
        conservative (bool, optional): Only run the most conservative diagnostics. Defaults to False.

    Returns:
        Dict[str, np.ndarray]: Dictionary of the the diagnostics results.
    """
    if conservative:
        return {
            "Distant": distant_diagnostic(sm, max_distance),
            "Heavyweight": heavyweight_diagnostic(sm, min_size),
            "Lightweight": lightweight_diagnostic(sm, max_size),
            "Weight Neighbourhood": weight_neighbourhood_diagnostic(
                sm, min_size, max_size
            ),
        }
    else:
        return {
            "Distant": distant_diagnostic(sm, max_distance),
            "Heavyweight": heavyweight_diagnostic(sm, min_size),
            "Lightweight": lightweight_diagnostic(sm, max_size),
            "Weight Neighbourhood": weight_neighbourhood_diagnostic(
                sm, min_size, max_size
            ),
            "Loss Neighbourhood": loss_neighbourhood_diagnostic(sm, min_size),
            "Global Loss": global_loss_diagnostic(sm),
            "Clustering": optics_diagnostic(sm, min_size),
        }



"""
This module contains the built-in alternatives for local white box models.
These can also be used as templates for implementing your own.
"""

from typing import Optional, Union

import numpy as np
import torch
from torch.nn.functional import softmax

from slisemap.utils import _assert


def linear_regression(X: torch.Tensor, B: torch.Tensor) -> torch.Tensor:
    """Prediction function for linear regression.

    Args:
        X (torch.Tensor): Data matrix [n_x, m].
        B (torch.Tensor): Coefficient Matrix [n_b, m].

    Returns:
        torch.Tensor: Prediction tensor [n_b, n_x, 1]
    """
    return (B @ X.T)[:, :, None]


def multiple_linear_regression(X: torch.Tensor, B: torch.Tensor) -> torch.Tensor:
    """Prediction function for multiple linear regression.

    Args:
        X (torch.Tensor): Data matrix [n_x, m].
        B (torch.Tensor): Coefficient Matrix [n_b, m*p].

    Returns:
        torch.Tensor: Prediction tensor [n_b, n_x, p]
    """
    n_x, m = X.shape
    n_b, o = B.shape
    p = torch.div(o, m, rounding_mode="trunc")
    a = torch.empty([n_b, n_x, p], device=B.device, dtype=B.dtype)
    for i in range(p):
        a[:, :, i] = B[:, (i * m) : ((i + 1) * m)] @ X.T
    return a


def linear_regression_loss(
    Ytilde: torch.Tensor, Y: torch.Tensor, B: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """Least squares loss function for (multiple) linear regresson.

    Args:
        Ytilde (torch.Tensor): Predicted values [n_b, n_x, p].
        Y (torch.Tensor): Ground truth values [n_x, p].
        B (Optional[torch.Tensor], optional): Coefficient matrix (not used, the regularisation is part of Slisemap). Defaults to None.

    Returns:
        torch.Tensor: Loss values [n_b, n_x].
    """
    return ((Ytilde - Y.expand(Ytilde.shape)) ** 2).sum(dim=-1)


def linear_regression_coefficients(
    X: Union[torch.Tensor, np.ndarray],
    Y: Union[torch.Tensor, np.ndarray],
    intercept: bool = False,
) -> int:
    """Get the number of coefficients for a (multiple) linear regression.

    Args:
        X (Union[torch.Tensor, np.ndarray]): Data matrix.
        Y (Union[torch.Tensor, np.ndarray]): Target matrix.
        intercept (bool, optional): Add an (additional) intercept to X. Defaults to False.

    Returns:
        int: Number of coefficients (columns of B).
    """
    return (X.shape[1] + intercept) * (1 if len(Y.shape) < 2 else Y.shape[1])


def logistic_regression(X: torch.Tensor, B: torch.Tensor) -> torch.Tensor:
    """Prediction function for (multinomial) logistic regression.
    Note that the number of coefficients is `m * (p-1)` due to the normalisation of softmax.

    Args:
        X (torch.Tensor): Data matrix [n_x, m].
        B (torch.Tensor): Coefficient Matrix [n_b, m*(p-1)].

    Returns:
        torch.Tensor: Prediction tensor [n_b, n_x, p]
    """
    n_x, m = X.shape
    n_b, o = B.shape
    p = 1 + torch.div(o, m, rounding_mode="trunc")
    a = torch.zeros([n_b, n_x, p], device=B.device, dtype=B.dtype)
    for i in range(p - 1):
        a[:, :, i] = B[:, (i * m) : ((i + 1) * m)] @ X.T
    return softmax(a, 2)


def logistic_regression_loss(
    Ytilde: torch.Tensor, Y: torch.Tensor, B: Optional[torch.Tensor] = None
) -> torch.Tensor:
    """Squared Hellinger distance function for (multinomial) logistic regression.

    Args:
        Ytilde (torch.Tensor): Predicted values [n_b, n_x, p].
        Y (torch.Tensor): Ground truth values [n_x, p].
        B (Optional[torch.Tensor], optional): Coefficient matrix (not used, the regularisation is part of Slisemap). Defaults to None.

    Returns:
        torch.Tensor: Loss values [n_b, n_x].
    """
    return ((Ytilde.sqrt() - Y.sqrt().expand(Ytilde.shape)) ** 2).sum(dim=-1) * 0.5


def logistic_regression_coefficients(
    X: Union[torch.Tensor, np.ndarray],
    Y: Union[torch.Tensor, np.ndarray],
    intercept: bool = False,
) -> int:
    """Get the number of coefficients for a (multinomial) logistic regression.

    Args:
        X (Union[torch.Tensor, np.ndarray]): Data matrix.
        Y (Union[torch.Tensor, np.ndarray]): Target matrix.
        intercept (bool, optional): Add an (additional) intercept to X. Defaults to False.

    Returns:
        int: Number of coefficients (columns of B).
    """
    _assert(
        len(Y.shape) > 1 and Y.shape[1] > 1,
        "Logistic regression requires Y:s with multiple classes",
    )
    return (X.shape[1] + intercept) * (Y.shape[1] - 1)



"""
This module contains various useful functions.
"""

from timeit import default_timer as timer
from typing import Any, Callable, List, Optional, Sequence, Union
from warnings import warn

import numpy as np
import torch


class SlisemapException(Exception):
    # Custom Exception type (for filtering)
    pass


class SlisemapWarning(Warning):
    # Custom Warning type (for filtering)
    pass


def _assert(condition: bool, message: str):
    if not condition:
        raise SlisemapException(message)


def _warn(warning: str, method: Optional[Callable] = None):
    if method is None:
        warn(warning, SlisemapWarning, 2)
    else:
        warn(f"{method.__qualname__}: {warning}", SlisemapWarning, 2)


def _tonp(x: torch.Tensor) -> np.ndarray:
    """Convert a torch.Tensor to a numpy.ndarray.

    Args:
        x (torch.Tensor): Input torch.Tensor.

    Returns:
        np.ndarray: Output numpy.ndarray.
    """
    return x.cpu().detach().numpy()


class CheckConvergence:
    """An object that tries to estimate when an optimisation has converged.
    Use it for, e.g., escape+optimisation cycles in Slisemap.
    """

    __slots__ = ("current", "best", "counter", "patience", "optimal_state")

    def __init__(self, patience: int = 3):
        """Create a CheckConvergence object.

        Args:
            patience (int, optional): How long should the optimisation continue without improvement. Defaults to 3.
        """
        self.current = np.inf
        self.best = np.inf
        self.counter = 0
        self.patience = max(1e-8, patience)
        self.optimal_state = None

    def has_converged(
        self,
        loss: Union[float, Sequence[float]],
        store: Optional[Callable[[], Any]] = None,
    ) -> bool:
        """Check if the optimisation has converged.

        Args:
            loss (Union[float, Sequence[float]]): The latest loss value(s).
            store (Optional[Callable[[], Any]], optional): Function that stores the optimal state in self.optimal_state. Defaults to None.


        Returns:
            bool: True if the optimisation has converged.
        """
        loss = np.asarray(loss)
        if np.any(np.isnan(loss)):
            _warn("Loss is `nan`", CheckConvergence.has_converged)
            return True
        if np.any(loss < self.best):
            self.best = np.minimum(loss, self.best)
            self.current = loss
            self.counter = 0  # Reset the counter if a new best
            if store is not None and loss.item(0) < self.best.item(0):
                self.optimal_state = store()
            return False
        # Increase the counter if no improvement
        self.counter += np.mean(self.current <= loss)
        self.current = loss
        return self.counter >= self.patience  # Has the patience run out


def LBFGS(
    loss_fn: Callable[[], torch.Tensor],
    variables: List[torch.Tensor],
    max_iter: int = 500,
    max_eval: Optional[int] = None,
    line_search_fn: Optional[str] = "strong_wolfe",
    time_limit: Optional[float] = None,
    **kwargs,
) -> torch.optim.LBFGS:
    """Optimise a function using LBFGS.

    Args:
        loss_fn (Callable[[], torch.Tensor]): Function that returns a value to be minimised.
        variables (List[torch.Tensor]): List of variables to optimise (must have `requires_grad=True`).
        max_iter (int, optional): Maximum number of LBFGS iterations. Defaults to 500.
        max_eval (Optional[int], optional): Maximum number of function evaluations. Defaults to `1.25 * max_iter`.
        line_search_fn (Optional[str], optional): Line search method (None or "strong_wolfe"). Defaults to "strong_wolfe".
        time_limit (Optional[float], optional): Optional time limit for the optimisation (in seconds). Defaults to None.
        **kwargs (optional): Argumemts passed to `torch.optim.LBFGS`.

    Returns:
        torch.optim.LBFGS: The LBFGS optimiser.
    """
    optimiser = torch.optim.LBFGS(
        variables,
        max_iter=max_iter if time_limit is None else 20,
        max_eval=max_eval,
        line_search_fn=line_search_fn,
        **kwargs,
    )

    def closure():
        optimiser.zero_grad()
        loss = loss_fn()
        loss.backward()
        return loss

    if time_limit is None:
        optimiser.step(closure)
    else:
        start = timer()
        prev_evals = 0
        for _ in range((max_iter - 1) // 20 + 1):
            optimiser.step(closure)
            if timer() - start > time_limit:
                break  # Time limit exceeded
            tot_evals = optimiser.state_dict()["state"][0]["func_evals"]
            if prev_evals + 1 == tot_evals:
                break  # LBFGS has converged if it returns after one evaluation
            prev_evals = tot_evals
            if max_eval is not None:
                if tot_evals >= max_eval:
                    break  # Number of evaluations exceeded max_eval
                optimiser.param_groups[0]["max_eval"] -= tot_evals
            # The number of steps is limited by ceiling(max_iter/20) with 20 iterations per step

    return optimiser


def PCA_rotation(
    X: torch.Tensor, components: int, full: bool = True, niter: int = 10
) -> torch.Tensor:
    """Calculate the rotation matrix from PCA.

    If the PCA fails (e.g. if original matrix is not full rank) then this shows a warning instead of throwing an error (returns a dummy rotation).

    Args:
        X (torch.Tensor): The original matrix.
        components (int): The number of components in the embedding.
        full (bool, optional): Use a full SVD for the PCA (slower). Defaults to True.
        niter (int, optional): The number of iterations when a randomised approach is used. Defaults to 10.

    Returns:
        torch.Tensor: Rotation matrix that turns the original matrix into the embedded space.
    """
    try:
        components = min(*X.shape, components)
        if full:
            return torch.linalg.svd(X, full_matrices=False)[2][:, :components]
        else:
            return torch.pca_lowrank(X, components, center=False, niter=niter)[2]
    except:
        _warn("Could not perform PCA", PCA_rotation)
        z = torch.zeros((X.shape[1], components), dtype=X.dtype, device=X.device)
        z.fill_diagonal_(1.0, True)
        return z


def varimax(
    X: torch.Tensor, gamma: float = 1.0, max_iter: float = 100, tolerance: float = 1e-8
) -> torch.Tensor:
    """Rotate a matrix using varimax (so that the first dimension has the largest variance).

    Code adapted from: http://en.wikipedia.org/wiki/Talk:Varimax_rotation

    Args:
        X (torch.Tensor): Matrix
        gamma (float, optional): Learning rate. Defaults to 1.0.
        max_iter (float, optional): Maximum number of iterations. Defaults to 100.
        tolerance (float, optional): Early stopping tolerance (relative). Defaults to 1e-8.

    Returns:
        torch.Tensor: Rotated matrix.
    """
    gamma /= X.shape[0]
    eye = torch.eye(X.shape[1], dtype=X.dtype, device=X.device)
    R = eye.clone()
    d = 0.0
    for _ in range(max_iter):
        XR = X @ R
        u, s, vh = torch.linalg.svd(
            X.T @ (XR**3 - gamma * (XR @ (eye * (XR.T @ XR)))), full_matrices=False
        )
        R = u @ vh
        d_old = d
        d = torch.sum(s)
        if d_old != 0.0 and d / d_old < 1.0 + tolerance:
            break
    return X @ R


def global_model(
    X: torch.Tensor,
    Y: torch.Tensor,
    local_model: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    local_loss: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
    coefficients: Optional[int] = None,
    lasso: float = 0.0,
    ridge: float = 0.0,
) -> torch.Tensor:
    """Find coefficients for a global model.

    Args:
        X (torch.Tensor): Data matrix.
        Y (torch.Tensor): Target matrix.
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Prediction function for the model.
        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]): Loss function for the model.
        coefficients (Optional[int], optional): Number of coefficients. Defaults to X.shape[1].
        lasso (float, optional): Lasso-regularisation coefficient for B ($\\lambda_{lasso} * ||B||_1$). Defaults to 0.0.
        ridge (float, optional): Ridge-regularisation coefficient for B ($\\lambda_{ridge} * ||B||_2$). Defaults to 0.0.

    Returns:
        torch.Tensor: Global model coefficients.
    """
    shape = (1, X.shape[1] * Y.shape[1] if coefficients is None else coefficients)
    B = torch.zeros(shape, dtype=X.dtype, device=X.device).requires_grad_(True)

    def loss():
        l = local_loss(local_model(X, B), Y, B).mean()
        if lasso > 0:
            l += lasso * torch.sum(B.abs())
        if ridge > 0:
            l += ridge * torch.sum(B**2)
        return l

    LBFGS(loss, [B])
    return B.detach()



"""
    This module contains functions that can be used to evaluate SLISEMAP solutions.
    The functions take a solution (plus other arguments) and returns a single float.
    This float should either be minimised or maximised for best results (see individual functions).
"""

from typing import Callable, Optional, Sequence, Tuple, Union

import numpy as np
import torch

from slisemap.slisemap import Slisemap
from slisemap.utils import _tonp


def _non_crashing_median(x: torch.Tensor) -> float:
    if x.numel() == 0:
        return np.nan
    else:
        return torch.median(x).cpu().item()


def nanmean(x) -> float:
    mask = np.isfinite(x)
    if np.all(mask):
        return np.mean(x)
    elif not np.any(mask):
        return np.nan
    else:
        return np.mean(x[mask])


def euclidean_nearest_neighbours(
    D: torch.Tensor, index: int, k: Union[int, float] = 0.1, include_self: bool = True
) -> torch.LongTensor:
    """Find the k nearest neighbours using euclidean distance

    Args:
        D (torch.Tensor): Distance matrix.
        index (int): The item (row), for which to find neighbours.
        k (int, optional): The number of neighbours to find. Defaults to 0.1.
        include_self (bool, optional): include the item in its' neighbourhood. Defaults to True.

    Returns:
        torch.LongTensor: Vector of indices for the neighbours.
    """
    if isinstance(k, float) and 0 < k <= 1:
        k = int(k * D.shape[0])
    if include_self:
        return torch.argsort(D[index])[:k]
    else:
        dist = D[index].clone()
        if not include_self:
            dist[index] = np.inf
            if k == D.shape[0]:
                k -= 1
        return torch.argsort(dist)[:k]


def kernel_neighbours(
    D: torch.Tensor, index: int, epsilon: float = 1.0, include_self: bool = True
) -> torch.LongTensor:
    """Find the neighbours using a softmax kernel

    Args:
        D (torch.Tensor): Distance matrix.
        index (int): The item for which we want to find neighbours.
        epsilon (float, optional): Treshold for selecting the neighbourhood (will be divided by `n`). Defaults to 1.0.
        include_self (bool, optional): include the item in its' neighbourhood. Defaults to True.

    Returns:
        torch.LongTensor: Vector of indices for the neighbours.
    """
    K = torch.nn.functional.softmax(-D[index], 0)
    epsilon2 = epsilon / K.numel()
    if include_self:
        return torch.where(K >= epsilon2)[0]
    else:
        mask = K >= epsilon2
        mask[index] = False
        return torch.where(mask)[0]


def cluster_neighbours(
    D: torch.Tensor,
    index: int,
    clusters: torch.LongTensor,
    include_self: bool = True,
) -> torch.LongTensor:
    """Find the neighbours with given clusters

    Args:
        D (torch.Tensor): Distance matrix (ignored).
        index (int): The item for which we want to find neighbours.
        clusters (torch.LongTensor): Cluster id:s for the data items.
        include_self (bool, optional): include the item in its' neighbourhood. Defaults to True.

    Returns:
        torch.LongTensor: Vector of indices for the neighbours.
    """
    if include_self:
        return torch.where(clusters == clusters[index])[0]
    else:
        mask = clusters == clusters[index]
        mask[index] = False
        return torch.where(mask)[0]


def radius_neighbours(
    D: torch.Tensor,
    index: int,
    radius: Optional[float] = None,
    quantile: float = 0.2,
    include_self: bool = True,
) -> torch.LongTensor:
    """Find the neighbours within a radius

    Args:
        D (torch.Tensor): Distance matrix (ignored).
        index (int): The item for which we want to find neighbours.
        radius (Optional[float], optional): The radius of the neighbourhood. Defaults to None.
        quantile (float, optional): If radius is None then radius is set to the quantile of D. Defaults to 0.2.
        include_self (bool, optional): include the item in its' neighbourhood. Defaults to True.

    Returns:
        torch.LongTensor: Vector of indices for the neighbours.
    """
    if radius is None:
        radius = torch.quantile(D, quantile)
    if include_self:
        return torch.where(D[index] <= radius)[0]
    else:
        mask = D[index] <= radius
        mask[index] = False
        return torch.where(mask)[0]


def _get_neighbours(
    sm: Union[Slisemap, torch.Tensor],
    neighbours: Union[
        None,
        np.ndarray,
        torch.LongTensor,
        Callable[[torch.Tensor, int], torch.LongTensor],
    ],
    full_if_none: bool = False,
    **kwargs,
) -> Callable[[int], torch.LongTensor]:
    """Create a function that takes the index of an item and returns the indices of its neighbours.

    The neighbours parameter is primarily one of:
     - euclidean_nearest_neighbours
     - kernel_neighbours
     - cluster_neighbours
     - radius_neighbours

    Args:
        sm (Union[Slisemap, torch.Tensor]): Trained Slisemap solution or an embedding vector (like Slisemap.Z).
        neighbours (Union[ None, np.ndarray, torch.LongTensor, Callable[[torch.Tensor, int], torch.LongTensor], ]): Either None (return self), a vector of cluster id:s (take neighbours from the same cluter), or a function that gives neighbours (that takes a distance matrix and an index).
        full_if_none (bool, optional): If `neighbours` is None, return the whole dataset. Defaults to False.
        **kwargs: Arguments passed on to `neighbours` (if it is a function).

    Returns:
        Callable[[int], torch.LongTensor]: Function that takes an index and returns neighbour indices.
    """
    if neighbours is None:
        if full_if_none:
            n = sm.n if isinstance(sm, Slisemap) else sm.shape[0]
            return lambda i: torch.arange(n)
        else:
            return lambda i: torch.LongTensor([i])
    if callable(neighbours):
        D = sm.get_D(numpy=False) if isinstance(sm, Slisemap) else torch.cdist(sm, sm)
        return lambda i: neighbours(D, i, **kwargs)
    else:
        neighbours = torch.as_tensor(neighbours)
        return lambda i: cluster_neighbours(None, i, neighbours, **kwargs)


def slisemap_loss(sm: Slisemap) -> float:
    """Evaluate a SLISEMAP solution by calculating the loss.

    Smaller is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.

    Returns:
        float: The loss value.
    """
    return sm.value()


def slisemap_entropy(sm: Slisemap) -> float:
    """Evaluate a SLISEMAP solution by calculating the entropy.

    Args:
        sm (Slisemap): Trained Slisemap solution.

    Returns:
        float: The loss value.
    """
    return sm.entropy()


def fidelity(
    sm: Slisemap,
    neighbours: Union[
        None,
        np.ndarray,
        torch.LongTensor,
        Callable[[torch.Tensor, int], torch.LongTensor],
    ] = None,
    **kwargs,
) -> float:
    """Evaluate a SLISEMAP solution by calculating the fidelity (loss per item/neighbourhood).

    Smaller is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        neighbours (Union[ None, np.ndarray, torch.LongTensor, Callable[[torch.Tensor, int], torch.LongTensor], ]): Either None (only corresponding local model), a vector of cluster id:s, or a function that gives neighbours (see `_get_neighbours`).
        **kwargs: Arguments passed on to `neighbours` (if it is a function).

    Returns:
        float: The mean loss.
    """
    neighbours = _get_neighbours(sm, neighbours, full_if_none=False, **kwargs)
    results = np.zeros(sm.n)
    L = sm.get_L(numpy=False)
    for i in range(len(results)):
        ni = neighbours(i)
        if ni.numel() == 0:
            results[i] = np.nan
        else:
            results[i] = torch.mean(L[i, ni]).cpu().detach().item()
    return nanmean(results)


def coverage(
    sm: Slisemap,
    max_loss: float,
    neighbours: Union[
        None,
        np.ndarray,
        torch.LongTensor,
        Callable[[torch.Tensor, int], torch.LongTensor],
    ] = None,
    **kwargs,
) -> float:
    """Evaluate a SLISEMAP solution by calculating the coverage.

    Larger is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        max_loss (float): Maximum tolerable loss.
        neighbours (Union[ None, np.ndarray, torch.LongTensor, Callable[[torch.Tensor, int], torch.LongTensor], ], optional): Either None (all), a vector of cluster id:s, or a function that gives neighbours (see `_get_neighbours`).
        **kwargs: Arguments passed on to `neighbours` (if it is a function).

    Returns:
        float: The mean fraction of items within the error bound.
    """
    if torch.all(torch.isnan(sm.B.sum(1))).cpu().item():
        return np.nan
    neighbours = _get_neighbours(sm, neighbours, full_if_none=True, **kwargs)
    results = np.zeros(sm.n)
    L = sm.get_L(numpy=False)
    for i in range(len(results)):
        ni = neighbours(i)
        if ni.numel() == 0:
            results[i] = np.nan
        else:
            results[i] = np.mean(_tonp(L[i, ni] < max_loss))
    return nanmean(results)


def median_loss(
    sm: Slisemap,
    neighbours: Union[
        None,
        np.ndarray,
        torch.LongTensor,
        Callable[[torch.Tensor, int], torch.LongTensor],
    ] = None,
    **kwargs,
) -> float:
    """Evaluate a SLISEMAP solution by calculating the median loss.

    Smaller is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        neighbours (Union[ None, np.ndarray, torch.LongTensor, Callable[[torch.Tensor, int], torch.LongTensor], ], optional): Either None (all), a vector of cluster id:s, or a function that gives neighbours (see `_get_neighbours`).
        **kwargs: Arguments passed on to `neighbours` (if it is a function).

    Returns:
        float: The mean median loss.
    """
    neighbours = _get_neighbours(sm, neighbours, full_if_none=True, **kwargs)
    results = np.zeros(sm.n)
    L = sm.get_L(numpy=False)
    for i in range(len(results)):
        ni = neighbours(i)
        if ni.numel() == 0:
            results[i] = np.nan
        else:
            results[i] = _non_crashing_median(L[i, ni])
    return nanmean(results)


def coherence(
    sm: Slisemap,
    neighbours: Union[
        None,
        np.ndarray,
        torch.LongTensor,
        Callable[[torch.Tensor, int], torch.LongTensor],
    ] = None,
    **kwargs,
) -> float:
    """Evaluate a SLISEMAP solution by calculating the coherence (max change in prediction divided by the change in variable values).

    Smaller is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        neighbours (Union[ None, np.ndarray, torch.LongTensor, Callable[[torch.Tensor, int], torch.LongTensor], ], optional): Either None (all), a vector of cluster id:s, or a function that gives neighbours (see `_get_neighbours`).
        **kwargs: Arguments passed on to `neighbours` (if it is a function).

    Returns:
        float: The mean coherence.
    """
    neighbours = _get_neighbours(
        sm, neighbours, full_if_none=True, include_self=False, **kwargs
    )
    results = np.zeros(sm.n)
    P = sm.local_model(sm.X, sm.B)
    for i in range(len(results)):
        ni = neighbours(i)
        if ni.numel() == 0:
            results[i] = np.nan
        else:
            dP = torch.sum((P[None, i, i] - P[ni, i] - P[i, ni] + P[ni, ni]) ** 2, 1)
            dX = torch.sum((sm.X[None, i, :] - sm.X[ni, :]) ** 2, 1) + 1e-8
            results[i] = torch.sqrt(torch.max(dP / dX))
    return nanmean(results)


def stability(
    sm: Slisemap,
    neighbours: Union[
        None,
        np.ndarray,
        torch.LongTensor,
        Callable[[torch.Tensor, int], torch.LongTensor],
    ] = None,
    **kwargs,
) -> float:
    """Evaluate a SLISEMAP solution by calculating the stability (max change in the local model divided by the change in variable values).

    Smaller is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        neighbours (Union[ None, np.ndarray, torch.LongTensor, Callable[[torch.Tensor, int], torch.LongTensor], ], optional): Either None (all), a vector of cluster id:s, or a function that gives neighbours (see `_get_neighbours`).
        **kwargs: Arguments passed on to `neighbours` (if it is a function).

    Returns:
        float: The mean stability.
    """
    neighbours = _get_neighbours(
        sm, neighbours, full_if_none=True, include_self=False, **kwargs
    )
    results = np.zeros(sm.n)
    for i in range(len(results)):
        ni = neighbours(i)
        if ni.numel() == 0:
            results[i] = np.nan
        else:
            dB = torch.sum((sm.B[None, i, :] - sm.B[ni, :]) ** 2, 1)
            dX = torch.sum((sm.X[None, i, :] - sm.X[ni, :]) ** 2, 1) + 1e-8
            results[i] = torch.sqrt(torch.max(dB / dX))
    return nanmean(results)


def kmeans_matching(
    sm: Slisemap, clusters: Union[int, Sequence[int]] = range(2, 10)
) -> float:
    """Evaluate SLISE by measuring how well clusters in Z and B overlap (using kmeans to find the clusters).
    The overlap is measured by finding the best matching clusters and dividing the size of intersect by the size of the union of each cluster pair.

    Larger is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        clusters (Union[int, Sequence[int]], optional): The number of clusters. Defaults to range(2, 10).

    Returns:
        float: mean cluster matching.
    """
    from sklearn.cluster import KMeans
    from scipy.optimize import linear_sum_assignment

    Z = sm.get_Z()
    B = sm.get_B()
    if np.all(np.var(Z, 0) < 1e-8) or np.all(np.var(B, 0) < 1e-8):
        return np.nan  # Do not compare singular clusters
    if not np.all(np.isfinite(Z)) or not np.all(np.isfinite(B)):
        return np.nan
    if isinstance(clusters, int):
        clusters = range(clusters, clusters + 1)
    results = []
    for k in clusters:
        cl_B = KMeans(n_clusters=k, init="k-means++").fit(B)
        cl_Z = KMeans(n_clusters=k, init="k-means++").fit(Z)
        sets_B = [set(np.where(cl_B.labels_ == i)[0]) for i in range(k)]
        sets_Z = [set(np.where(cl_Z.labels_ == i)[0]) for i in range(k)]
        mat = np.zeros((k, k))
        for i, sB in enumerate(sets_B):
            for j, sZ in enumerate(sets_Z):
                mat[i, j] = len(sB.intersection(sZ)) / (len(sB.union(sZ)) + 1e-8)
        # Hungarian algorithm to find the best match between the clusterings
        rows, cols = linear_sum_assignment(mat, maximize=True)
        results.append(mat[rows, cols].mean())
    return nanmean(results)


def cluster_purity(
    sm: Union[Slisemap, torch.Tensor, np.ndarray],
    clusters: Union[np.ndarray, torch.LongTensor],
) -> float:
    """Evaluate a SLISEMAP solution by calculating how many items in the same cluster are neighbours.

    Larger is better.

    Args:
        sm (Union[Slisemap, torch.Tensor, np.ndarray]): Trained Slisemap solution _or_ embedding matrix.
        clusters (Union[np.ndarray, torch.LongTensor]): Cluster ids.

    Returns:
        float: The mean number of items sharing cluster that are neighbours.
    """
    if isinstance(sm, Slisemap):
        Z = sm.get_Z(numpy=False)
    elif isinstance(sm, torch.Tensor):
        Z = sm
    else:
        Z = torch.as_tensor(sm)
    if isinstance(clusters, np.ndarray):
        clusters = torch.as_tensor(clusters, device=Z.device)
    res = np.zeros(Z.shape[0])
    D = torch.cdist(Z, Z)
    for i in range(len(res)):
        mask = clusters[i] == clusters
        knn = euclidean_nearest_neighbours(D, i, mask.sum())
        res[i] = torch.sum(mask[knn]) / knn.shape[0]
    return nanmean(res)


def kernel_purity(
    sm: Slisemap,
    clusters: Union[np.ndarray, torch.LongTensor],
    epsilon: float = 1.0,
    losses: bool = False,
) -> float:
    """Evaluate a SLISEMAP solution by calculating how many neighbours are in the same cluster

    Larger is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        clusters (Union[np.ndarray, torch.LongTensor]): Cluster ids.
        epsilon (float, optional): Treshold for being a neighbour (`softmax(D) < epsilon/n`). Defaults to 1.0.
        losses (bool, optional): Use losses instead of embedding distances. Defaults to False.

    Returns:
        float: The mean number of neighbours that are in the same cluster.
    """
    if isinstance(clusters, np.ndarray):
        clusters = torch.tensor(clusters)
    res = np.zeros(sm.n)
    if losses:
        D = sm.get_L(numpy=False)
    else:
        D = sm.get_D(numpy=False)
    for i in range(len(res)):
        mask = clusters[i] == clusters
        neig = kernel_neighbours(D, i, epsilon)
        res[i] = torch.sum(mask[neig]) / neig.numel()
    return nanmean(res)


def recall(sm: Slisemap, epsilon_D: float = 1.0, epsilon_L: float = 1.0) -> float:
    """Evaluate a SLISEMAP solution by calculating the recall.

    We define recall as the intersection between the loss and embedding neighbourhoods divided by the loss neighbourhood.

    Larger is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        epsilon_D (float, optional): Treshold for being an embedding neighbour (`softmax(D) < epsilon/n`). Defaults to 1.0.
        epsilon_L (float, optional): Treshold for being a loss neighbour (`softmax(L) < epsilon/n`). Defaults to 1.0.

    Returns:
        float: The mean recall.
    """
    res = np.zeros(sm.n)
    D = sm.get_D(numpy=False)
    L = sm.get_L(numpy=False)
    for i in range(len(res)):
        nL = kernel_neighbours(L, i, epsilon_L)
        if nL.numel() == 0:
            res[i] = np.nan
        else:
            nD = kernel_neighbours(D, i, epsilon_D)
            inter = np.intersect1d(_tonp(nD), _tonp(nL), True)
            res[i] = len(inter) / nL.numel()
    return nanmean(res)


def precision(sm: Slisemap, epsilon_D: float = 1.0, epsilon_L: float = 1.0) -> float:
    """Evaluate a SLISEMAP solution by calculating the recall.

    We define recall as the intersection between the loss and embedding neighbourhoods divided by the embedding neighbourhood.

    Larger is better.

    Args:
        sm (Slisemap): Trained Slisemap solution.
        epsilon_D (float, optional): Treshold for being an embedding neighbour (`softmax(D) < epsilon/n`). Defaults to 1.0.
        epsilon_L (float, optional): Treshold for being a loss neighbour (`softmax(L) < epsilon/n`). Defaults to 1.0.

    Returns:
        float: The mean precision.
    """
    res = np.zeros(sm.n)
    D = sm.get_D(numpy=False)
    L = sm.get_L(numpy=False)
    for i in range(len(res)):
        nD = kernel_neighbours(D, i, epsilon_D)
        if nD.numel() == 0:
            res[i] = np.nan
        else:
            nL = kernel_neighbours(L, i, epsilon_L)
            inter = np.intersect1d(_tonp(nD), _tonp(nL), True)
            res[i] = len(inter) / nD.numel()
    return nanmean(res)


def relevance(sm: Slisemap, pred_fn: Callable, change: float) -> float:
    """Evaluate a SLISEMAP solution by calculating the relevance.

    Smaller is better.

    TODO: This does not (currently) work for multi-class predictions

    Args:
        sm (Slisemap): Trained Slisemap solution.
        pred_fn (Callable): Function that gives y:s for new x:s (the "black box model").
        change (float): How much should the prediction change?

    Returns:
        float: The mean number of mutated variables required to cause a large enough change in the prediction.
    """
    rel = np.ones(sm.n) * sm.m
    for i in range(len(rel)):
        b = sm.B[i, :]
        x = sm.X[i, :]
        y = sm.Y[i, 0]
        xmax = torch.max(sm.X, 0)[0]
        xmin = torch.min(sm.X, 0)[0]
        xinc = torch.where(b > 0, xmax, xmin)
        xdec = torch.where(b < 0, xmax, xmin)
        babs = torch.abs(b)
        for i, bs in enumerate(torch.sort(babs)[0]):
            yinc = pred_fn(torch.where(babs >= bs, xinc, x))
            ydec = pred_fn(torch.where(babs >= bs, xdec, x))
            if yinc - y > change or y - ydec > change:
                rel[i] = i
                break
    return nanmean(rel)


def accuracy(
    sm: Slisemap,
    X: Optional[np.ndarray] = None,
    Y: Optional[np.ndarray] = None,
    **kwargs,
) -> float:
    """Evaluate a SLISEMAP solution by checking how well the fitted models work on new points

    Args:
        sm (Slisemap): Trained Slisemap solution.
        X (Optional[np.ndarray], optional): New data matrix (uses the training data if None). Defaults to None.
        Y (Optional[np.ndarray], optional): New target matrix (uses the training data if None). Defaults to None.
        **kwargs: Optional keyword arguments to Slisemap.fit_new.

    Returns:
        float: Mean loss for the new points.
    """
    if X is None or Y is None:
        X = sm.get_X(intercept=False, numpy=False)
        Y = sm.get_Y(numpy=False)
    B, Z, loss = sm.fit_new(X, Y, loss=True, **kwargs)
    return loss.mean()



"""
This module contains alternative escape heuristics.
"""

from typing import Callable, Optional, Tuple

import numpy as np
import torch

from slisemap.utils import _assert


def escape_neighbourhood(
    X: torch.Tensor,
    Y: torch.Tensor,
    B: torch.Tensor,
    Z: torch.Tensor,
    local_model: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    local_loss: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
    distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    kernel: Callable[[torch.Tensor], torch.Tensor],
    radius: float = 3.5,
    force_move: bool = False,
    **_,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Try to escape a local optimum by moving the data items.
    Move the data items to the neighbourhoods (embedding and local model) best suited for them.
    This is done by finding another item (in the optimal neighbourhood) and copying its values for Z and B.

    Args:
        X (torch.Tensor): Data matrix.
        Y (torch.Tensor): Target matrix.
        B (torch.Tensor): Local models.
        Z (torch.Tensor): Embedding matrix.
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Prediction function for the local models.
        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]): Loss function for the local models.
        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Embedding distance function.
        kernel (Callable[[torch.Tensor], torch.Tensor]):  Kernel for embedding distances.
        radius (float, optional): For enforcing the radius of Z. Defaults to 3.5.
        force_move (bool, optional): Do not allow the items to pair with themselves. Defaults to True.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Escaped `B` and `Z`.
    """
    L = local_loss(local_model(X, B), Y, B)
    if radius > 0:
        Z2 = Z * (radius / (torch.sqrt(torch.sum(Z**2) / Z.shape[0]) + 1e-8))
        D = distance(Z2, Z2)
    else:
        D = distance(Z, Z)
    W = kernel(D)
    # K = torch.zeros_like(L)
    # for i in range(L.shape[1]):
    #     K[:, i] = torch.sum(W * L[:, i].ravel()[None, :], 1)
    # K = torch.sum(W[:, :, None] * L[None, :, :], 1)
    K = W @ L
    if force_move:
        _assert(
            K.shape[0] == K.shape[1],
            "force_move only works if (X, Y) corresponds to (B, Z)",
        )
        K.fill_diagonal_(np.inf)
    index = torch.argmin(K, 0)
    return B.detach()[index].clone(), Z.detach()[index].clone()


def escape_greedy(
    X: torch.Tensor,
    Y: torch.Tensor,
    B: torch.Tensor,
    Z: torch.Tensor,
    local_model: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    local_loss: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
    distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    kernel: Callable[[torch.Tensor], torch.Tensor],
    radius: float = 3.5,
    force_move: bool = False,
    **_,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Try to escape a local optimum by moving the data items.
    Move the data items to a locations with optimal local models.
    This is done by finding another item (with an optimal local model) and copying its values for Z and B.

    Args:
        X (torch.Tensor): Data matrix.
        Y (torch.Tensor): Target matrix.
        B (torch.Tensor): Local models.
        Z (torch.Tensor): Embedding matrix.
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Prediction function for the local models.
        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]): Loss function for the local models.
        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Embedding distance function.
        kernel (Callable[[torch.Tensor], torch.Tensor]):  Kernel for embedding distances.
        radius (float, optional): For enforcing the radius of Z. Defaults to 3.5.
        force_move (bool, optional): Do not allow the items to pair with themselves. Defaults to True.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Escaped `B` and `Z`.
    """
    L = local_loss(local_model(X, B), Y, B)
    if force_move:
        _assert(
            L.shape[0] == L.shape[1],
            "force_move only works if (X, Y) corresponds to (B, Z)",
        )
        L.fill_diagonal_(np.inf)
    index = torch.argmin(L, 0)
    return B.detach()[index].clone(), Z.detach()[index].clone()


def escape_combined(
    X: torch.Tensor,
    Y: torch.Tensor,
    B: torch.Tensor,
    Z: torch.Tensor,
    local_model: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    local_loss: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
    distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    kernel: Callable[[torch.Tensor], torch.Tensor],
    radius: float = 3.5,
    force_move: bool = False,
    **_,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Try to escape a local optimum by moving the data items.
    Move the data items to the neighbourhoods (embedding and local model) best suited for them.
    This is done by finding another item (in the optimal neighbourhood) and copying its values for Z and B.

    This is a combination of escape_neighbourhood and escape_greedy.

    Args:
        X (torch.Tensor): Data matrix.
        Y (torch.Tensor): Target matrix.
        B (torch.Tensor): Local models.
        Z (torch.Tensor): Embedding matrix.
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Prediction function for the local models.
        local_loss (Callable[[torch.Tensor, torch.Tensor], torch.Tensor, torch.Tensor]): Loss function for the local models.
        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Embedding distance function.
        kernel (Callable[[torch.Tensor], torch.Tensor]):  Kernel for embedding distances.
        radius (float, optional): For enforcing the radius of Z. Defaults to 3.5.
        force_move (bool, optional): Do not allow the items to pair with themselves. Defaults to True.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Escaped `B` and `Z`.
    """
    L = local_loss(local_model(X, B), Y, B)
    if radius > 0:
        Z2 = Z * (radius / (torch.sqrt(torch.sum(Z**2) / Z.shape[0]) + 1e-8))
        D = distance(Z2, Z2)
    else:
        D = distance(Z, Z)
    W = kernel(D)
    K = W @ L + L
    if force_move:
        _assert(
            K.shape[0] == K.shape[1],
            "force_move only works if (X, Y) corresponds to (B, Z)",
        )
        K.fill_diagonal_(np.inf)
    index = torch.argmin(K, 0)
    return B.detach()[index].clone(), Z.detach()[index].clone()


def escape_marginal(
    X: torch.Tensor,
    Y: torch.Tensor,
    B: torch.Tensor,
    Z: torch.Tensor,
    local_model: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    local_loss: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
    distance: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    kernel: Callable[[torch.Tensor], torch.Tensor],
    radius: float = 3.5,
    force_move: bool = False,
    Xold: Optional[torch.Tensor] = None,
    Yold: Optional[torch.Tensor] = None,
    jit: bool = True,
    **_,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Try to escape a local optimum by moving the data items.
    Move the data items to locations with optimal marginal losses.
    This is done by finding another item (where the marginal loss is optimal) and copying its values for Z and B.

    This might produce better results than `escape_neighbourhood`, but is really slow.

    Args:
        X (torch.Tensor): Data matrix.
        Y (torch.Tensor): Target matrix.
        B (torch.Tensor): Local models.
        Z (torch.Tensor): Embedding matrix.
        local_model (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Prediction function for the local models.
        local_loss (Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor]): Loss function for the local models.
        distance (Callable[[torch.Tensor, torch.Tensor], torch.Tensor]): Embedding distance function.
        kernel (Callable[[torch.Tensor], torch.Tensor]):  Kernel for embedding distances.
        radius (float, optional): For enforcing the radius of Z. Defaults to 3.5.
        force_move (bool, optional): Do not allow the items to pair with themselves. Defaults to True.
        jit (bool, optional): Just-In-Time compile the loss function. Defaults to True.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Escaped `B` and `Z`.
    """
    from slisemap.slisemap import make_marginal_loss

    _assert(
        not force_move or X.shape[0] == Z.shape[0],
        "force_move only works if (X, Y) corresponds to (B, Z)",
    )
    _assert(
        Xold is not None or X.shape[0] == Z.shape[0],
        "(Xold Yold) is required if (X, Y) does not correspond to (B, Z)",
    )
    if radius > 0:
        Z2 = Z * (radius / (torch.sqrt(torch.sum(Z**2) / Z.shape[0]) + 1e-8))
    else:
        Z2 = Z
    lf, set_new = make_marginal_loss(
        X=X if Xold is None else Xold,
        Y=Y if Yold is None else Yold,
        B=B,
        Z=Z2,
        Xnew=X[:1],
        Ynew=Y[:1],
        local_model=local_model,
        local_loss=local_loss,
        distance=distance,
        kernel=kernel,
        radius=0.0,
        lasso=0.0,
        ridge=0.0,
        jit=jit and (X.shape[0] > 10 if Xold is None else Xold.shape[0] > 10),
    )
    index = []
    for i in range(X.shape[0]):
        set_new(X[i : i + 1], Y[i : i + 1])
        index.append(i)
        best = np.inf
        for j in range(Z.shape[0]):
            if force_move and i == j:
                continue
            l = lf(B[j : j + 1], Z2[j : j + 1])
            if l < best:
                best = l
                index[-1] = j
    return B.detach()[index].clone(), Z.detach()[index].clone()
